#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for `scopus.AffiliationRetrieval` module."""

from nose.tools import assert_equal, assert_true

from pybliometrics.scopus import AffiliationRetrieval


light = AffiliationRetrieval('60000356', refresh=30, view="LIGHT")
standard = AffiliationRetrieval('60000356', refresh=30, view="STANDARD")


def test_address():
    assert_equal(light.address, 'Private Bag X3')
    assert_equal(standard.address, 'Private Bag X3')


def test_affiliation_name():
    assert_equal(light.affiliation_name, 'University of Cape Town')
    assert_equal(standard.affiliation_name, 'University of Cape Town')


def test_author_count():
    expected = 12800
    assert_true(light.author_count >= expected)
    assert_true(standard.author_count >= expected)


def test_city():
    assert_equal(light.city, 'Cape Town')
    assert_equal(standard.city, 'Cape Town')


def test_country():
    assert_equal(light.country, 'South Africa')
    assert_equal(standard.country, 'South Africa')


def test_date_created():
    assert_equal(light.date_created, None)
    assert_equal(standard.date_created, (2008, 2, 2))


def test_document_count():
    expected = 73581
    assert_true(light.document_count >= expected)
    assert_true(standard.document_count >= expected)


def test_eid():
    assert_equal(light.eid, '10-s2.0-60000356')
    assert_equal(standard.eid, '10-s2.0-60000356')


def test_identifier():
    assert_equal(light.identifier, 60000356)
    assert_equal(standard.identifier, 60000356)


def test_name_variants():
    expected = "<class 'pybliometrics.scopus.affiliation_retrieval.Variant'>"
    assert_equal(str(type(light.name_variants[0])), expected)
    assert_equal(str(type(standard.name_variants[0])), expected)


def test_org_domain():
    assert_equal(light.org_domain, None)
    assert_equal(standard.org_domain, 'uct.ac.za')


def test_org_type():
    assert_equal(light.org_type, None)
    assert_equal(standard.org_type, 'univ')


def test_org_URL():
    assert_equal(light.org_URL, None)
    assert_equal(standard.org_URL, 'http://www.uct.ac.za')


def test_postal_code():
    assert_equal(light.postal_code, None)
    assert_equal(standard.postal_code, '7701')


def test_scopus_affiliation_link():
    expected = 'https://www.scopus.com/affil/profile.uri?afid='\
               '60000356&partnerID=HzOxMe3b&origin=inward'
    assert_equal(light.scopus_affiliation_link, expected)
    assert_equal(standard.scopus_affiliation_link, expected)


def test_self_link():
    expected = 'https://api.elsevier.com/content/affiliation/affiliation_id/60000356'
    assert_equal(light.self_link, expected)
    assert_equal(standard.self_link, expected)


def test_search_link():
    expected = 'https://api.elsevier.com/content/search/scopus?query=af-id%2860000356%29'
    assert_equal(light.search_link, expected)
    assert_equal(standard.search_link, expected)


def test_state():
    assert_equal(light.state, None)
    assert_equal(standard.state, 'Western Cape')


def test_status():
    assert_equal(light.status, None)
    assert_equal(standard.status, "update")


def sort_name():
    assert_equal(light.sort_name, None)
    assert_equal(standard.sort_name, 'Cape Town, University of')


def url():
    expected = 'https://api.elsevier.com/content/affiliation/affiliation_id/60000356'
    assert_equal(light.url, expected)
    assert_equal(standard.url, expected)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for `scopus.SerialSearch` module."""

from nose.tools import assert_equal, assert_true

from pybliometrics.scopus import SerialSearch


# Search by title
ser1 = SerialSearch({'title': 'SoftwareX'}, refresh=30)
# Search by ISSN
ser2 = SerialSearch({'issn': '1468-0262'}, refresh=30)
# Search by publisher
ser3 = SerialSearch({'pub': 'Stellenbosch'}, refresh=30)
# Search by subject abbreviation
ser4 = SerialSearch({'subj': 'COMP'}, refresh=30)
# Search by subject area code
ser5 = SerialSearch({'subjCode': '2612'}, refresh=30)


def test_results_title():
    assert_equal(len(ser1.results), 1)
    assert_equal(ser1.results[0]['title'], 'SoftwareX')


def test_results_issn():
    assert_equal(len(ser2.results), 1)
    assert_equal(ser2.results[0]['title'], 'Econometrica')


def test_results_pub():
    assert_equal(len(ser3.results), 5)
    assert_equal(ser3.results[0]['title'], 'African Finance Journal')


def test_results_subj():
    ser4_subj_abbs = set(i['subject_area_abbrevs'] for i in ser4.results)
    assert_true(False not in ['COMP' in i for i in ser4_subj_abbs])


def test_results_subjcode():
    ser5_subj_codes = set(i['subject_area_codes'] for i in ser5.results)
    assert_true(False not in ['2612' in i for i in ser5_subj_codes])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for `scopus.ScopusSearch` module."""

from collections import namedtuple

from nose.tools import assert_equal, assert_true

from pybliometrics.scopus import ScopusSearch

order = 'eid doi pii pubmed_id title subtype subtypeDescription creator '\
        'afid affilname  affiliation_city affiliation_country author_count '\
        'author_names author_ids author_afids coverDate coverDisplayDate '\
        'publicationName issn source_id eIssn aggregationType volume '\
        'issueIdentifier article_number pageRange description authkeywords '\
        'citedby_count openaccess freetoread freetoreadLabel fund_acr '\
        'fund_no fund_sponsor'
doc = namedtuple('Document', order)

# Set to refresh=False because of citation count
s_au = ScopusSearch('AU-ID(24320488600)', refresh=30)
s_j = ScopusSearch('SOURCE-ID(22900) AND PUBYEAR IS 2010', refresh=30)
q_empty = 'SOURCE-ID(19700188323) AND PUBYEAR IS 1900'
s_empty = ScopusSearch(q_empty, refresh=30)


def test_get_eids_author():
    expected = ['2-s2.0-85117005558', '2-s2.0-84937325266', '2-s2.0-26444452434']
    assert_equal(s_au.get_eids(), expected)


def test_get_eids_journal():
    assert_equal(len(s_j.get_eids()), 118)


def test_get_results_size():
    assert_equal(s_au.get_results_size(), 3)
    assert_equal(s_j.get_results_size(), 118)
    assert_equal(s_empty.get_results_size(), 0)


def test_results_author():
    received = s_au.results[-1]
    expected = doc(eid='2-s2.0-26444452434', doi='10.1016/0014-2921(92)90085-B',
        pii='001429219290085B', pubmed_id=None,
        title='Economists as policymakers: A round-table discussion. Introduction',
        subtype='ar', subtypeDescription='Article', creator='Draghi M.',
        afid=None, affilname=None, affiliation_city=None, affiliation_country=None,
        author_count='1', author_names='Draghi, Mario', author_ids='24320488600',
        author_afids=None, coverDate='1992-01-01', coverDisplayDate='April 1992',
        publicationName='European Economic Review', issn='00142921',
        source_id='20749', eIssn=None, aggregationType='Journal', volume='36',
        issueIdentifier='2-3', article_number=None, pageRange='307-309',
        description=None, authkeywords=None, citedby_count=0, openaccess=0,
        freetoread=None, freetoreadLabel=None, fund_acr=None,
        fund_no='undefined', fund_sponsor=None)
    assert_true(int(received.citedby_count) > 0)
    assert_equal(received._replace(citedby_count=0), expected)


def test_results_journal():
    received = s_j.results[-1]
    abstract = 'This paper investigates the determinants of R&D investment '\
        'at the national level with an emphasis on the roles of patent '\
        'rights protection, international technology transfer through trade '\
        'and FDI, and economic growth, in addition to the essentials of '\
        'human capital accumulation and the number of scientific '\
        'researchers. The Extreme-Bounds-Analysis (EBA) approach is applied '\
        'to examine the robustness and sensitivity of these factors. The '\
        'results of the EBA tests on data from 26 OECD countries from 1996 '\
        'to 2006 showed that tertiary education and the proportion of '\
        'scientific researchers in a country were robust determinants that '\
        'had positive effects on R&D intensity. Foreign technology inflows '\
        'had a robust and negative impact on domestic R&D. Patent rights '\
        'protection and the income growth rate are fragile determinants of '\
        'R&D investment. © 2009 Elsevier B.V. All rights reserved.'
    keywords = 'Extreme-Bounds-Analysis (EBA) | Patent rights protection | '\
        'R&amp;D investment | Technology transfer'
    title = 'Determinants of R&amp;D investment: The Extreme-Bounds-'\
            'Analysis approach applied to 26 OECD countries'
    expected = doc(eid='2-s2.0-74249121335', doi='10.1016/j.respol.2009.11.010',
        pii='S0048733309002145', pubmed_id=None, title=title, subtype='ar',
        subtypeDescription='Article', creator='Wang E.', afid='60007954',
        affilname='National Chung Cheng University',
        affiliation_city='Min-Hsiung', affiliation_country='Taiwan',
        author_count='1', author_names='Wang, Eric C.',
        author_ids='7403414138', author_afids='60007954',
        coverDate='2010-01-01', coverDisplayDate='2010',
        publicationName='Research Policy', issn='00487333', source_id='22900',
        eIssn=None, aggregationType='Journal', volume='39', issueIdentifier='1',
        article_number=None, pageRange='103-116', description=abstract,
        authkeywords=keywords, citedby_count=0, openaccess=0, freetoread=None,
        freetoreadLabel=None, fund_acr='MOST', fund_no='NSC94-2415-H-194-001',
        fund_sponsor='Ministry of Science and Technology, Taiwan')
    assert_true(int(received.citedby_count) > 60)
    assert_equal(received._replace(citedby_count=0), expected)


def test_results_empty():
    assert_equal(s_empty.results, None)



from pybliometrics.scopus.utils.checks import *
from pybliometrics.scopus.utils.constants import *
from pybliometrics.scopus.utils.create_config import *
from pybliometrics.scopus.utils.get_content import *
from pybliometrics.scopus.utils.parse_content import *
from pybliometrics.scopus.utils.startup import *



from collections import OrderedDict
from typing import Dict, List, Optional, Union

from pybliometrics.scopus.superclasses import Search
from pybliometrics.scopus.utils import check_parameter_value, make_search_summary


class SerialSearch(Search):
    @property
    def results(self) -> Optional[List[Dict[str, str]]]:
        """A list of OrderedDicts representing results of serial search. The
        number of keys may vary from one search result to another depending
        on the length of yearly data.
        """
        out = []
        search_results = self._json['serial-metadata-response'].get('entry', [])
        for result in search_results:
            # OrderedDict to populate with individual serial data
            obs = OrderedDict()
            for key, value in result.items():
                if not value:
                    continue
                key = key.split(":", 1)[-1]
                if key == '@_fa':
                    continue
                elif key == 'subject-area':
                    subject_data = _merge_subject_data(value)
                    obs['subject_area_codes'] = subject_data[0]
                    obs['subject_area_abbrevs'] = subject_data[1]
                    obs['subject_area_names'] = subject_data[2]
                elif key == 'SNIPList' or key == 'SJRList':
                    for j in _retrieve_source_rankings(value):
                        obs[j[0]] = j[1]
                elif key == 'citeScoreYearInfoList':
                    for j in _retrieve_cite_scores(value):
                        obs[j[0]] = j[1]
                elif key == 'link':
                    for j in _retrieve_links(value):
                        obs[j[0]] = j[1]
                elif key == 'yearly-data':
                    time_data = _retrieve_yearly_data(value.get('info', []))
                    for j in time_data:
                        obs[j[0]] = j[1]
                else:
                    obs[key] = value
            if obs:
                out.append(obs)
        return out or None

    def __init__(self,
                 query: Dict,
                 refresh: Union[bool, int] = False,
                 view: str = 'ENHANCED'
                 ) -> None:
        """Interaction with the Serial Title API.

        :param query:  Query parameters and corresponding fields. Allowed keys
                      'title', 'issn', 'pub', 'subj', 'subjCode', 'content',
                      'oa'.  For examples on possible values, please refer to
                      https://dev.elsevier.com/documentation/SerialTitleAPI.wadl#d1e22.
        :param refresh: Whether to refresh the cached file if it exists or not.
                        If int is passed, cached file will be refreshed if the
                        number of days since last modification exceeds that value.
        :param view: The view of the file that should be downloaded.  Allowed
                     values: STANDARD, ENHANCED, CITESCORE.  For details see
                     https://dev.elsevier.com/sc_serial_title_views.html.

        Raises
        ------
        Scopus400Error
            If provided value for a query key is invalid or if for
            non-subscribers the number of search results exceeds 5000.

        ValueError
            If any of the parameters `refresh` or `view` is not
            one of the allowed values.

        Notes
        -----
        The directory for cached results is `{path}/{view}/{fname}`,
        where `path` is specified in your configuration file, and `fname` is
        the md5-hashed version of `query` dict turned into string in format
        of 'key=value' delimited by '&'.
        """
        # Checks
        allowed_query_keys = ('title', 'issn', 'date', 'pub', 'subj',
                              'subjCode', 'content', 'oa')
        invalid = [k for k in query.keys() if k not in allowed_query_keys]
        if invalid:
            raise ValueError(f'Query key(s) "{", ".join(invalid)}" invalid')
        check_parameter_value(view, ('STANDARD', 'ENHANCED', 'CITESCORE'), "view")

        # Query
        self._query = str(query)
        self._refresh = refresh
        self._view = view
        Search.__init__(self, query=query, api='SerialSearch')
        self._n = len(self._json['serial-metadata-response'].get('entry', []))

    def __str__(self):
        """Print a summary string."""
        titles = [d['title'] for d in self.results]
        return make_search_summary(self, "source", titles)


def _merge_subject_data(subject_area_data):
    """Auxiliary function to collect and concatenate subject area data into string.

    Returns tuple of strings for subject area names, subject area codes and
    subject area abbreviations deliminated by ';'.
    """
    codes = set([j.get('@code') for j in subject_area_data])
    abbrevs = set([j.get('@abbrev') for j in subject_area_data])
    names = set([j.get('$') for j in subject_area_data])
    return (';'.join(filter(None, codes)),
            ';'.join(filter(None, abbrevs)),
            ';'.join(filter(None, names)))


def _retrieve_links(link_data):
    """Auxiliary function to collect data on links.

    Returns list of lists in the form of [linkname, link].
    """
    out = []
    for sl in link_data:
        try:
            out.append([sl['@ref'], sl.get('@href')])
        except KeyError:
            continue
    return out or None


def _retrieve_yearly_data(yearly_data):
    """Auxiliary function to collect yearly data.

    Returns list of lists in the form [mergedstatname, stat], where
    mergedstatname - dictionary key joined with associated period.
    """
    out = []
    for t in yearly_data:
        for key in t:
            if key not in ('@year', '@_fa'):
                out.append([f"{key}_{t['@year']}", t[key]])
    return out or None


def _retrieve_cite_scores(cite_score_data):
    """Auxiliary function to collect citescore data.

    Returns list of lists in the form [mergedstatname, stat], where
    mergedstatname - dictionary key joined with associated period.
    """
    out = []
    for key in ("citeScoreTracker", "citeScoreCurrentMetric"):
        try:
            year = f"{key}_{cite_score_data[key + 'Year']}"
            val = cite_score_data[key]
            out.append([year, val])
        except KeyError:
            pass
    return out or None


def _retrieve_source_rankings(source_data):
    """Auxiliary function to collect SNIP and SJR data.

    Returns list of lists in the form [mergedstatname, stat],
    where mergedstatname - dictionary key joined with associated period
    """
    out = []
    for key in source_data:
        stats = source_data[key]
        for t in stats:
            out.append([f"{key}_['@year']", t['$']])
    return out or None



from os import environ
from pathlib import Path

# Paths for cached files
if (Path.home()/".scopus").exists():
    BASE_PATH = Path.home()/".scopus"
else:
    BASE_PATH = Path.home()/".pybliometrics"/"Scopus"
DEFAULT_PATHS = {
    'AbstractRetrieval': BASE_PATH/'abstract_retrieval',
    'AffiliationRetrieval': BASE_PATH/'affiliation_retrieval',
    'AffiliationSearch': BASE_PATH/'affiliation_search',
    'AuthorRetrieval': BASE_PATH/'author_retrieval',
    'AuthorSearch': BASE_PATH/'author_search',
    'CitationOverview': BASE_PATH/'citation_overview',
    'ScopusSearch': BASE_PATH/'scopus_search',
    'SerialSearch': BASE_PATH/'serial_search',
    'SerialTitle': BASE_PATH/'serial_title',
    'PlumXMetrics': BASE_PATH/'plumx',
    'SubjectClassifications': BASE_PATH/'subject_classification'
}

# Configuration file location
if 'PYB_CONFIG_FILE' in environ:
    CONFIG_FILE = Path(environ['PYB_CONFIG_FILE'])
else:
    if (Path.home()/".scopus").exists():
        CONFIG_FILE = Path.home()/".scopus"/"config.ini"
    else:
        CONFIG_FILE = Path.home()/".pybliometrics"/"config.ini"

# URLs for all classes
RETRIEVAL_BASE = 'https://api.elsevier.com/content/'
SEARCH_BASE = 'https://api.elsevier.com/content/search/'
URLS = {
    'AbstractRetrieval': RETRIEVAL_BASE + 'abstract/',
    'AffiliationRetrieval': RETRIEVAL_BASE + 'affiliation/affiliation_id/',
    'AffiliationSearch': SEARCH_BASE + 'affiliation',
    'AuthorRetrieval': RETRIEVAL_BASE + 'author/author_id/',
    'AuthorSearch': SEARCH_BASE + 'author',
    'CitationOverview': RETRIEVAL_BASE + 'abstract/citations/',
    'ScopusSearch': SEARCH_BASE + 'scopus',
    'SerialSearch': RETRIEVAL_BASE + 'serial/title',
    'SerialTitle': RETRIEVAL_BASE + 'serial/title/issn/',
    'SubjectClassifications': RETRIEVAL_BASE + 'subject/scopus',
    'PlumXMetrics': 'https://api.elsevier.com/analytics/plumx/'
}

# Throttling limits (in queries per second) // 0 = no limit
RATELIMITS = {
    'AbstractRetrieval': 9,
    'AffiliationRetrieval': 9,
    'AffiliationSearch': 6,
    'AuthorRetrieval': 3,
    'AuthorSearch': 2,
    'CitationOverview': 4,
    'ScopusSearch': 9,
    'SerialSearch': 6,
    'SerialTitle': 6,
    'PlumXMetrics': 6,
    'SubjectClassifications': 0
}

# Other API restrictions
SEARCH_MAX_ENTRIES = 5_000



from collections import namedtuple
from typing import List, NamedTuple, Optional, Tuple, Union

from pybliometrics.scopus.superclasses import Retrieval
from pybliometrics.scopus.utils import chained_get, check_parameter_value,\
    get_link, make_int_if_possible


class SerialTitle(Retrieval):
    @property
    def aggregation_type(self) -> str:
        """The type of the source."""
        return self._entry['prism:aggregationType']

    @property
    def citescoreyearinfolist(self) -> Optional[List[Tuple[int, float, None]]]:
        """A list of two tuples of the form (year, cite-score).  The first
        tuple represents the current cite-score, the second tuple
        represents the tracker cite-score.  For discontinued sources, there 
        is no tracker cite-score, and the second tuple is (None, None).  See
        https://service.elsevier.com/app/answers/detail/a_id/30562/supporthub/scopus/.
        """
        try:
            d = self._entry['citeScoreYearInfoList']
        except KeyError:
            return None
        current = (int(d['citeScoreCurrentMetricYear']),
                   float(d['citeScoreCurrentMetric']))
        tracker = (d['citeScoreTrackerYear'], d['citeScoreTracker'])
        try:
            tracker = (int(tracker[0]), float(tracker[1]))
        except TypeError:
            pass
        return [current, tracker]

    @property
    def eissn(self) -> Optional[str]:
        """The electronic ISSN of the source."""
        return self._entry.get('prism:eIssn')

    @property
    def issn(self) -> Optional[str]:
        """The ISSN of the source."""
        return self._entry.get('prism:issn')

    @property
    def oaallowsauthorpaid(self) -> Optional[str]:
        """Whether under the Open-Access policy authors are allowed to pay."""
        return self._entry.get('oaAllowsAuthorPaid')

    @property
    def openaccess(self) -> Optional[int]:
        """Open Access status (0 or 1)."""
        return make_int_if_possible(chained_get(self._entry, ['openaccess']))

    @property
    def openaccessstartdate(self):
        """Starting availability date."""
        return self._entry.get('openaccessStartDate')

    @property
    def openaccesstype(self) -> Optional[str]:
        """Open Archive status (full or partial)."""
        return self._entry.get('openaccessType')

    @property
    def openaccessarticle(self) -> Optional[bool]:
        """Open Access status."""
        return self._entry.get('openaccessArticle')

    @property
    def openarchivearticle(self) -> Optional[bool]:
        """Open Archive status."""
        return self._entry.get('openArchiveArticle')

    @property
    def openaccesssponsorname(self) -> Optional[str]:
        """The name of the Open Access sponsor."""
        return self._entry.get('openaccessSponsorName')

    @property
    def openaccesssponsortype(self) -> Optional[str]:
        """The type of the Open Access sponsor."""
        return self._entry.get('openaccessSponsorType')

    @property
    def openaccessuserlicense(self) -> Optional[str]:
        """The User license."""
        return self._entry.get('openaccessUserLicense')

    @property
    def publisher(self) -> str:
        """The publisher of the source."""
        return self._entry['dc:publisher']

    @property
    def scopus_source_link(self) -> str:
        """URL to info site on scopus.com."""
        return get_link(self._entry, 0, ["link"])

    @property
    def self_link(self) -> str:
        """URL to the source's API page."""
        return get_link(self._json, 0, ["link"])

    @property
    def sjrlist(self) -> Optional[List[Tuple[int, float]]]:
        """The SCImago Journal Rank (SJR) indicator as list of
        (year, indicator)-tuples.  See
        https://www.scimagojr.com/journalrank.php.
        """
        return _parse_list(self._entry, "SJR")

    @property
    def sniplist(self) -> Optional[List[Tuple[int, float]]]:
        """The Source-Normalized Impact per Paper (SNIP) as list of
        (year, indicator)-tuples.  See
        https://blog.scopus.com/posts/journal-metrics-in-scopus-source-normalized-impact-per-paper-snip.
        """
        return _parse_list(self._entry, "SNIP")

    @property
    def source_id(self) -> int:
        """The Scopus ID of the source."""
        return int(self._entry['source-id'])

    @property
    def subject_area(self) -> Optional[List[NamedTuple]]:
        """List of named tuples of subject areas in the form
        (area, abbreviation, code) of the source.
        """
        area = namedtuple('Subjectarea', 'area abbreviation code')
        areas = [area(area=item['$'], code=int(item['@code']),
                      abbreviation=item['@abbrev'])
                 for item in self._entry["subject-area"]]
        return areas or None

    @property
    def title(self) -> str:
        """The title of the source."""
        return self._entry['dc:title']

    def __init__(self,
                 issn: Union[int, str],
                 refresh: Union[bool, int] = False,
                 view: str = "ENHANCED",
                 years: str = None
                 ) -> None:
        """Interaction with the Serial Title API.

        :param issn: The ISSN or the E-ISSN of the source.
        :param refresh: Whether to refresh the cached file if it exists or not.
                        If int is passed, cached file will be refreshed if the
                        number of days since last modification exceeds that value.
        :param view: The view of the file that should be downloaded.  Allowed
                     values: BASIC, STANDARD, ENHANCED.  For details see
                     https://dev.elsevier.com/sc_serial_title_views.html.
        :param years: A string specifying a year or range of years (combining
                      two years with a hyphen) for which yearly metric data
                      (SJR, SNIP, yearly-data) should be looked up for.  If
                      None, only the most recent metric data values are
                      provided. Note: If not None, refresh will always be True.

        Raises
        ------
        ValueError
            If any of the parameters `refresh` or `view` is not
            one of the allowed values.

        Notes
        -----
        The directory for cached results is `{path}/{view}/{source_id}`,
        where `path` is specified in your configuration file.
        """
        # Checks
        check_parameter_value(view, ('BASIC', 'STANDARD', 'ENHANCED'), "view")
        self._view = view

        # Force refresh when years is specified
        if years:
            refresh = True
        self._refresh = refresh

        # Load json
        self._id = str(issn)
        self._years = years
        Retrieval.__init__(self, identifier=self._id, date=years,
                           api='SerialTitle')

        # Parse json
        self._json = self._json['serial-metadata-response']
        self._entry = self._json['entry'][0]

    def __str__(self):
        """Print a summary string."""
        date = self.get_cache_file_mdate().split()[0]
        areas = [e.area for e in self.subject_area]
        if len(areas) == 1:
            areas = areas[0]
        else:
            areas = " and ".join([", ".join(areas[:-1]), areas[-1]])
        s = f"'{self.title}', {self.aggregation_type} published by "\
            f"'{self.publisher}', is active in {areas}\n"
        metrics = []
        if self.sjrlist:
            metrics.append(f"SJR:  year value")
            for rec in self.sjrlist:
                metrics.append(f"      {rec[0]} {rec[1]}")
        if self.sniplist:
            metrics.append(f"SNIP: year value")
            for rec in self.sniplist:
                metrics.append(f"      {rec[0]} {rec[1]}")
        if metrics:
            s += f"Metrics as of {date}:\n    " + "\n    ".join(metrics) + "\n"
        s += f"    ISSN: {self.issn or '-'}, E-ISSN: {self.eissn or '-'}, "\
             f"Scopus ID: {self.source_id}"
        return s


def _parse_list(d, metric):
    """Auxiliary function to parse SNIP and SJR lists."""
    try:
        values = [(int(r['@year']), float(r['$'])) for r
                  in d[metric + "List"][metric]]
        return sorted(set(values))
    except (KeyError, TypeError):
        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for `scopus.AffiliationSearch` module."""

from collections import namedtuple

from nose.tools import assert_equal, assert_true

from pybliometrics.scopus import AffiliationSearch

s1 = AffiliationSearch('AF-ID(60021784)', refresh=30)
s2 = AffiliationSearch('AFFIL(Max Planck Munich)', download=False, refresh=True)


def test_affiliations():
    received1 = s1.affiliations
    assert_true(isinstance(received1, list))
    order = 'eid name variant documents city country parent'
    Affiliation = namedtuple('Affiliation', order)
    expected = Affiliation(eid='10-s2.0-60021784', name='New York University',
        variant='', documents=0, city='New York', country='United States',
        parent='0')
    assert_true(received1[0].documents >= 90_000)
    assert_equal(received1[0]._replace(documents=0), expected)


def test_affiliations_nodownload():
    assert_equal(s2.affiliations, None)


def test_get_results_size():
    received1 = s1.get_results_size()
    assert_true(received1 >= 1)
    received2 = s2.get_results_size()
    assert_true(received2 >= 60)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for `scopus.AbstractRetrieval` module."""

from collections import namedtuple
from nose.tools import assert_equal, assert_true

from pybliometrics.scopus import AbstractRetrieval

# Base information
ab1 = AbstractRetrieval("2-s2.0-84930616647", view="FULL", refresh=30)
# Conference proceeding and no references
ab2 = AbstractRetrieval("2-s2.0-0029486824", view="FULL", refresh=30)
# Issuetitle and no affiliation
ab3 = AbstractRetrieval("2-s2.0-0001270077", view="FULL", refresh=30)
# Author group broken and author keywords
ab4 = AbstractRetrieval("2-s2.0-0000016206", view="FULL", refresh=30)
# ISBN
ab5 = AbstractRetrieval("2-s2.0-84919546381", view="FULL", refresh=30)
# Funding, sequencebanks, chemicals
ab6 = AbstractRetrieval("2-s2.0-85040230676", view="FULL", refresh=30)
# Contributor group
ab7 = AbstractRetrieval("2-s2.0-85050253030", view="FULL", refresh=30)
# REF view
ab8 = AbstractRetrieval("2-s2.0-84951753303", view="REF", refresh=30)


def test_abstract():
    expected = 'In this paper we propose a Bayesian analysis of seasonal '\
        'unit roots in quarterly observed time series. Seasonal unit root '\
        'processes are useful to describe economic series with changing '\
        'seasonal fluctuations. A natural alternative model for similar '\
        'purposes contains deterministic seasonal mean shifts instead of '\
        'seasonal stochastic trends. This leads to analysing seasonal unit '\
        'roots in the presence of mean shifts using Bayesian techniques. '\
        'Our method is illustrated using several simulated and empirical data.'
    assert_equal(ab4.abstract, expected)
    assert_equal(ab8.abstract, None)


def test_affiliation():
    aff = namedtuple('Affiliation', 'id name city country')
    expected = [aff(id=60027950, name='Carnegie Mellon University',
                    city='Pittsburgh', country='United States')]
    assert_equal(ab1.affiliation, expected)
    assert_equal(ab3.affiliation, None)
    assert_equal(ab8.affiliation, None)


def test_aggregationType():
    assert_equal(ab1.aggregationType, 'Journal')
    assert_equal(ab2.aggregationType, 'Conference Proceeding')
    assert_equal(ab8.aggregationType, None)


def test_authkeywords():
    assert_equal(ab1.authkeywords, None)
    expected = ['Bayesian analysis', 'Seasonality',
                'Structural breaks', 'Unit roots']
    assert_equal(ab4.authkeywords, expected)
    assert_equal(ab8.authkeywords, None)


def test_authorgroup():
    fields = 'affiliation_id dptid organization city postalcode '\
             'addresspart country auid orcid indexed_name surname given_name'
    auth = namedtuple('Author', fields)
    expected = [auth(affiliation_id=60027950, dptid=110785688,
        organization='Department of Chemical Engineering, Carnegie Mellon University',
        city='Pittsburgh', postalcode='15213', addresspart='5000 Forbes Avenue',
        country='United States', auid=7004212771, orcid=None,
        indexed_name='Kitchin J.', surname='Kitchin', given_name='John R.')]
    assert_equal(ab1.authorgroup, expected)
    assert_equal(ab8.authorgroup, None)


def test_authors():
    fields = 'auid indexed_name surname given_name affiliation_id'
    auth = namedtuple('Author', fields)
    expected = [auth(auid=7004212771, indexed_name='Kitchin J.R.',
                surname='Kitchin', given_name='John R.',
                affiliation_id='60027950')]
    assert_equal(ab1.authors, expected)
    assert_equal(ab8.authors, None)


def test_citedby_count():
    expected = 5
    assert_true(ab1.citedby_count >= expected)
    assert_equal(ab8.citedby_count, None)


def test_citedby_link():
    expected = 'https://www.scopus.com/inward/citedby.uri?partnerID=HzOxMe3b'\
        '&scp=84930616647&origin=inward'
    assert_equal(ab1.citedby_link, expected)
    assert_equal(ab8.citedby_link, None)


def test_chemials():
    received = ab6.chemicals
    assert_true(isinstance(received, list))
    assert_equal(len(received), 6)
    chemical = namedtuple('Chemical', 'source chemical_name cas_registry_number')
    expected = chemical(source='esbd', cas_registry_number='126547-89-5',
        chemical_name='intercellular adhesion molecule 1')
    assert_true(expected in received)
    assert_equal(ab3.chemicals, None)
    assert_equal(ab8.chemicals, None)


def test_confcode():
    assert_equal(ab2.confcode, 44367)
    assert_equal(ab8.confcode, None)


def test_confdate():
    assert_equal(ab2.confdate, ((1995, 12, 13), (1995, 12, 15)))
    assert_equal(ab8.confdate, None)


def test_conflocation():
    assert_equal(ab2.conflocation, 'New Orleans, LA, USA')
    assert_equal(ab8.conflocation, None)


def test_confname():
    expected2 = "Proceedings of the 1995 34th IEEE Conference on Decision "\
                "and Control. Part 1 (of 4)"
    assert_equal(ab2.confname, expected2)
    assert_equal(ab3.confname, None)
    expected7 = '20th Symposium on Design, Test, Integration and Packaging '\
                'of MEMS and MOEMS, DTIP 2018'
    assert_equal(ab7.confname, expected7)
    assert_equal(ab8.confname, None)


def test_confsponsor():
    assert_equal(ab2.confsponsor, 'IEEE')
    assert_equal(ab3.confsponsor, None)
    expected7 = ['ARTOV.IMM.CNR.IT', 'CMP.IMAG.FR', 'CNRS.FR',
                 'EPS.IEEE.ORG', 'LIRMM.FR']
    assert_equal(ab7.confsponsor, expected7)
    assert_equal(ab8.confsponsor, None)


def test_contributor_group():
    fields = 'given_name initials surname indexed_name role'
    pers = namedtuple('Contributor', fields)
    expected = pers(given_name='Romolo', initials='R.', surname='Marcelli',
                    indexed_name='Marcelli R.', role='edit')
    received = ab7.contributor_group
    assert_equal(len(received), 7)
    assert_true(expected in received)
    assert_equal(ab3.contributor_group, None)
    assert_equal(ab8.contributor_group, None)


def test_correspondence():
    fields = 'surname initials organization country city_group'
    corr = namedtuple('Correspondence', fields)
    expected2 = corr(surname='Boukas', initials='E.K.',
        organization='Ecole Polytechnique de Montreal', country='Canada',
        city_group='Montreal')
    assert_equal(ab2.correspondence[0], expected2)
    assert_equal(ab3.correspondence, None)
    assert_equal(ab8.correspondence, None)


def test_coverDate():
    assert_equal(ab1.coverDate, '2015-06-05')
    assert_equal(ab8.coverDate, None)


def test_description():
    expected = 'In this paper we propose a Bayesian analysis of seasonal '\
               'unit roots in quarterly observed time series. Seasonal unit '\
               'root processes are useful to describe economic series with '\
               'changing seasonal fluctuations. A natural alternative model '\
               'for similar purposes contains deterministic seasonal mean '\
               'shifts instead of seasonal stochastic trends. This leads to '\
               'analysing seasonal unit roots in the presence of mean '\
               'shifts using Bayesian techniques. Our method is illustrated '\
               'using several simulated and empirical data.'
    assert_equal(ab4.description, expected)
    assert_equal(ab8.description, None)


def test_doi():
    assert_equal(ab1.doi, '10.1021/acscatal.5b00538')
    assert_equal(ab4.doi, '10.1016/s0304-4076(97)00018-3')
    assert_equal(ab8.doi, None)


def test_eid():
    assert_equal(ab1.eid, '2-s2.0-84930616647')
    assert_equal(ab8.eid, None)


def test_endingPage():
    assert_equal(ab1.endingPage, '3899')
    assert_equal(ab8.endingPage, None)


def test_funding():
    received = ab6.funding
    assert_true(isinstance(received, list))
    assert_equal(len(received), 5)
    fund = namedtuple('Funding', 'agency agency_id string funding_id acronym country')
    expected6 = fund(
        agency='Deutsche Forschungsgemeinschaft',
        agency_id='http://data.elsevier.com/vocabulary/SciValFunders/501100001659',
        string='German Research Foundation', acronym='DFG',
        funding_id=['SFB685'],
        country='http://sws.geonames.org/2921044/')
    assert_true(expected6 in received)
    assert_equal(ab5.funding, None)
    assert_equal(ab8.funding, None)


def test_funding_text():
    e = 'ACKNOWLEDGMENTS. We thank Dieter Blaas for providing ICAM-1–specific '\
        'antiserum (supersup). This work was supported by Netherlands '\
        'Organization for Scientific Research Grant NWO-VICI-91812628 '\
        '(to F.J.M.v.K.), by German Research Foundation Grant SFB685 '\
        '(to T.S. and G.Z.), and Wellcome Trust PhD Studentship support '\
        '102572/B/13/Z (to D.L.H.). All EM was performed in the Astbury '\
        'Biostructure Laboratory, which was funded by the University of '\
        'Leeds and the Wellcome Trust (108466/Z/15/Z).'
    assert_equal(ab6.funding_text, e)
    assert_equal(ab8.funding_text, None)


def test_get_bibtex():
    e = '@article{Kaufmann1991FairnessPricing,\n  author = {Patrick J. '\
        'Kaufmann and Gwen Ortmeyer and N. Craig Smith},\n  title = {{'\
        'Fairness in consumer pricing}},\n  journal = {Journal of Consumer '\
        'Policy},\n  year = {1991},\n  volume = {14},\n  number = {2},\n  '\
        'pages = {117-140},\n  doi = {10.1007/BF00381915}}'
    assert_equal(ab3.get_bibtex(), e)


def test_get_latex():
    e = 'Philip Hans Franses, Henk Hoekh and Richard Paap, \\textit{'\
        'Bayesian analysis of seasonal unit roots and seasonal mean shifts}, '\
        'Journal of Econometrics, \\textbf{78(2)}, pp. 359-380 (1997). \\'\
        'href{https://doi.org/10.1016/s0304-4076(97)00018-3}{doi:10.1016/'\
        's0304-4076(97)00018-3}, \\href{https://www.scopus.com/inward/'\
        'record.uri?partnerID=HzOxMe3b&scp=0000016206&origin=inward}'\
        '{scopus:2-s2.0-0000016206}.'
    assert_equal(ab4.get_latex(), e)


def test_get_ris():
    e = 'TY  - JOUR\nTI  - Examples of effective data sharing in scientific '\
        'publishing\nJO  - ACS Catalysis\nVL  - 5\nDA  - 2015-06-05\nPY  - '\
        '2015\nSP  - 3894-3899\nAU  - Kitchin J.R.\nDO  - 10.1021/'\
        'acscatal.5b00538\nUR  - https://doi.org/10.1021/acscatal.5b00538\n'\
        'IS  - 6\nER  - \n\n'
    assert_equal(ab1.get_ris(), e)


def test_get_html():
    e = '<a href="https://www.scopus.com/authid/detail.url?origin=Author'\
        'Profile&authorId=7201922462">Patrick J. Kaufmann</a>, <a href="'\
        'https://www.scopus.com/authid/detail.url?origin=AuthorProfile&'\
        'authorId=16430389100">Gwen Ortmeyer</a> and <a href="https://'\
        'www.scopus.com/authid/detail.url?origin=AuthorProfile&authorId='\
        '57225963896">N. Craig Smith</a>, <a href="https://www.scopus.com/'\
        'inward/record.uri?partnerID=HzOxMe3b&scp=0001270077&origin=inward">'\
        'Fairness in consumer pricing</a>, <a href="https://www.scopus.com/'\
        'source/sourceInfo.url?sourceId=130073">Journal of Consumer Policy'\
        '</a>, <b>14(2)</b>, pp. 117-140, (1991). <a href="https://doi.org/'\
        '10.1007/BF00381915">doi:10.1007/BF00381915</a>.'
    assert_equal(ab3.get_html(), e)


def test_isbn():
    assert_equal(ab3.isbn, None)
    assert_equal(ab5.isbn, ('0203881486', '9780203881484'))
    assert_equal(ab8.isbn, None)


def test_issn():
    assert_equal(ab1.issn, '21555435')
    assert_equal(ab5.issn, None)
    assert_equal(ab8.issn, None)


def test_identifier():
    assert_equal(ab1.identifier, 84930616647)
    assert_equal(ab8.identifier, None)


def test_idxterms():
    expected = ['Control variables', 'Cost function',
                'Hamilton-Jacobi-Isaacs equation', 'Machine capacity',
                'Stochastic manufacturing systems', 'Value function']
    assert_equal(ab2.idxterms, expected)
    assert_equal(ab4.idxterms, None)


def test_issueIdentifier():
    assert_equal(ab1.issueIdentifier, '6')
    assert_equal(ab8.issueIdentifier, None)


def test_issuetitle():
    assert_equal(ab3.issuetitle, 'Law, Economics and Behavioural Sciences')
    assert_equal(ab8.issuetitle, None)


def test_language():
    assert_equal(ab1.language, 'eng')
    assert_equal(ab8.language, None)


def test_openaccess():
    assert_equal(ab5.openaccess, 2)
    assert_equal(ab6.openaccess, 1)
    assert_equal(ab7.openaccess, 0)
    assert_equal(ab8.openaccess, None)


def test_openaccessFlag():
    assert_equal(ab5.openaccessFlag, None)
    assert_equal(ab6.openaccessFlag, True)
    assert_equal(ab7.openaccessFlag, False)
    assert_equal(ab8.openaccessFlag, None)


def test_pageRange():
    assert_equal(ab1.pageRange, '3894-3899')
    assert_equal(ab8.pageRange, None)


def test_pii():
    assert_equal(ab4.pii, 'S0304407697000183')
    assert_equal(ab5.pii, None)
    assert_equal(ab6.pii, None)


def test_publicationName():
    assert_equal(ab1.publicationName, 'ACS Catalysis')
    assert_equal(ab8.publicationName, None)


def test_publisher():
    assert_equal(ab1.publisher, 'American Chemical Society')
    assert_equal(ab8.publisher, None)


def test_publisheraddress():
    assert_equal(ab2.publisheraddress, 'Piscataway, NJ, United States')
    assert_equal(ab8.publisheraddress, None)


def test_pubmed_id():
    assert_equal(ab6.pubmed_id, 29284752)
    assert_equal(ab7.pubmed_id, None)


def test_refcount():
    assert_equal(ab4.refcount, 18)
    assert_equal(ab8.refcount, 48)


def test_references_full():
    fields = 'position id doi title authors authors_auid authors_affiliationid '\
             'sourcetitle publicationyear volume issue first last citedbycount '\
             'type text fulltext'
    ref = namedtuple('Reference', fields)
    fulltext1 = 'Implementing Reproducible Research; Stodden, V.; Leisch, '\
                'F.; Peng, R. D., Eds., Chapman and Hall/CRC: London, 2014.'
    expected1 = ref(position='22', id='85055586929', doi=None, title=None,
        authors='Stodden, V.; Leisch, F.; Peng, R.D.', authors_auid=None,
        authors_affiliationid=None, fulltext=fulltext1,
        sourcetitle='Implementing Reproducible Research', type=None,
        publicationyear='2014', volume=None, issue=None, first=None,
        last=None, citedbycount=None, text='Eds. Chapman and Hall/CRC: London.')
    assert_equal(ab1.references[-1], expected1)
    assert_equal(ab2.references, None)


def test_references_ref():
    fields = 'position id doi title authors authors_auid authors_affiliationid '\
             'sourcetitle publicationyear volume issue first last citedbycount '\
             'type text fulltext'
    ref = namedtuple('Reference', fields)
    authors8 = 'Armbrust, Michael; Fox, Armando; Griffith, Rean; Joseph, '\
        'Anthony D.; Katz, Randy; Konwinski, Andy; Lee, Gunho; '\
        'Patterson, David; Rabkin, Ariel; Stoica, Ion; Zaharia, Matei'
    expected8 = ref(position='1', id='77950347409', authors=authors8,
        doi='10.1145/1721654.1721672', title='A view of cloud computing',
        sourcetitle='Communications of the ACM', type='resolvedReference',
        publicationyear=None, volume='53', issue='4', first='50',
        last='58', text=None, fulltext=None, citedbycount='0',
        authors_auid='35800975300; 35571093800; 57198081560; 57192787435; '\
            '7401788602; 25926395200; 56326032000; 7401930147; 26534952300; '\
            '7007009125; 15064891400',
        authors_affiliationid='60025038; 60025038; 60025038; 60025038; 60025038; '\
            '60025038; 60025038; 60025038; 60025038; 60025038; 60025038')
    assert_true(int(ab8.references[0].citedbycount) >= 0)
    assert_equal(ab8.references[0]._replace(citedbycount="0"), expected8)


def test_scopus_link():
    expected = 'https://www.scopus.com/inward/record.uri?partnerID=HzOxMe3b&'\
        'scp=84930616647&origin=inward'
    assert_equal(ab1.scopus_link, expected)


def test_self_link():
    expected = 'https://api.elsevier.com/content/abstract/scopus_id/84930616647'
    assert_equal(ab1.self_link, expected)
    assert_equal(ab8.self_link, None)


def test_sequencebank():
    received = ab6.sequencebank
    assert_true(isinstance(received, list))
    assert_equal(len(received), 3)
    bank = namedtuple('Chemical', 'name sequence_number type')
    expected = bank(name='GENBANK', sequence_number='MG272373',
                    type='referenced')
    assert_true(expected in received)
    assert_equal(ab3.sequencebank, None)
    assert_equal(ab8.sequencebank, None)


def test_source_id():
    assert_equal(ab1.source_id, 19700188320)
    assert_equal(ab8.source_id, None)


def test_sourcetitle_abbreviation():
    assert_equal(ab1.sourcetitle_abbreviation, 'ACS Catal.')
    assert_equal(ab8.sourcetitle_abbreviation, None)


def test_srctype():
    assert_equal(ab1.srctype, 'j')
    assert_equal(ab2.srctype, 'p')
    assert_equal(ab8.srctype, None)


def test_startingPage():
    assert_equal(ab1.startingPage, '3894')
    assert_equal(ab8.startingPage, None)


def test_subject_areas():
    area = namedtuple('Area', 'area abbreviation code')
    expected = [area(area='Catalysis', abbreviation='CENG', code=1503),
                area(area='Chemistry (all)', abbreviation='CHEM', code=1600)]
    assert_equal(ab1.subject_areas, expected)
    assert_equal(ab8.subject_areas, None)


def test_subtype():
    assert_equal(ab1.subtype, "re")
    assert_equal(ab2.subtype, "cp")
    assert_equal(ab5.subtype, "bk")
    assert_equal(ab6.subtype, "ar")


def test_subtypedescription():
    assert_equal(ab1.subtypedescription, "Review")
    assert_equal(ab2.subtypedescription, "Conference Paper")
    assert_equal(ab5.subtypedescription, "Book")
    assert_equal(ab6.subtypedescription, "Article")


def test_title():
    expected = 'Examples of effective data sharing in scientific publishing'
    assert_equal(ab1.title, expected)
    assert_equal(ab8.title, None)


def test_url():
    expected = 'https://api.elsevier.com/content/abstract/scopus_id/84930616647'
    assert_equal(ab1.url, expected)
    assert_equal(ab8.url, None)


def test_volume():
    assert_equal(ab1.volume, '5')
    assert_equal(ab8.volume, None)


def test_website():
    assert_equal(ab1.website, 'http://pubs.acs.org/page/accacs/about.html')
    assert_equal(ab2.website, None)
    assert_equal(ab8.website, None)



import requests

from pybliometrics import version_info
from pybliometrics.scopus import exception
from pybliometrics.scopus.utils.startup import config

# Define user agent string for HTTP requests
user_agent = 'pybliometrics-v' + '.'.join([str(e) for e in version_info[:3]])

errors = {400: exception.Scopus400Error, 401: exception.Scopus401Error,
          403: exception.Scopus403Error, 404: exception.Scopus404Error,
          413: exception.Scopus413Error, 414: exception.Scopus414Error,
          429: exception.Scopus429Error, 500: exception.Scopus500Error,
          502: exception.Scopus502Error}


def get_content(url, api, params={}, **kwds):
    """Helper function to download a file and return its content.

    Parameters
    ----------
    url : str
        The URL to be parsed.

    api : str
        The Scopus API to be accessed.

    params : dict (optional)
        Dictionary containing query parameters.  For required keys
        and accepted values see e.g.
        https://api.elsevier.com/documentation/AuthorRetrievalAPI.wadl

    **kwds : key-value parings, optional
        Keywords passed on to as query parameters.  Must contain fields
        and values specified in the respective API specification.

    Raises
    ------
    ScopusHtmlError or HTTPError
        If the status of the response is not ok.

    ValueError
        If the accept parameter is not one of the accepted values.

    Returns
    -------
    resp : byte-like object
        The content of the file, which needs to be serialized.
    """
    from random import shuffle
    from time import sleep, time

    from simplejson import JSONDecodeError

    from pybliometrics.scopus.utils.startup import _throttling_params, KEYS

    # Set header, params and proxy
    header = {'X-ELS-APIKey': KEYS[0],
              'Accept': 'application/json',
              'User-Agent': user_agent}
    if config.has_option('Authentication', 'InstToken'):
        token = config.get('Authentication', 'InstToken')
        header['X-ELS-Insttoken'] = token
    params.update(**kwds)
    proxies = dict(config._sections.get("Proxy", {}))

    # Eventually wait bc of throttling
    if len(_throttling_params[api]) == _throttling_params[api].maxlen:
        try:
            sleep(1 - (time() - _throttling_params[api][0]))
        except (IndexError, ValueError):
            pass

    # Perform request, eventually replacing the current key
    resp = requests.get(url, headers=header, proxies=proxies, params=params)
    while resp.status_code == 429:
        try:
            KEYS.pop(0)  # Remove current key
            shuffle(KEYS)
            header['X-ELS-APIKey'] = KEYS[0].strip()
            resp = requests.get(url, headers=header, proxies=proxies,
                                params=params)
        except IndexError:  # All keys depleted
            break
    _throttling_params[api].append(time())

    # Eventually raise error, if possible with supplied error message
    try:
        error_type = errors[resp.status_code]
        try:
            reason = resp.json()['service-error']['status']['statusText']
        except (JSONDecodeError, KeyError):
            try:
                reason = resp.json()['message']
            except:
                reason = ""
        raise errors[resp.status_code](reason)
    except KeyError:
        resp.raise_for_status()
    return resp


def detect_id_type(sid):
    """Method that tries to infer the type of abstract ID.

    Parameters
    ----------
    sid : str
        The ID of an abstract on Scopus.

    Raises
    ------
    ValueError
        If the ID type cannot be inferred.

    Notes
    -----
    Scopus IDs and Pubmed IDs are sometimes hard to distinguish.  If you
    work with both types, consider specifying the ID type manually.
    """
    sid = str(sid)
    if not sid.isnumeric():
        if sid.startswith('1-s2.0-') or sid.startswith('2-s2.0-'):
            id_type = 'eid'
        elif '/' in sid or "." in sid:
            id_type = 'doi'
        elif 16 <= len(sid) <= 17:
            id_type = 'pii'
    else:
        if len(sid) < 10:
            id_type = 'pubmed_id'
        else:
            id_type = 'scopus_id'
    try:
        return id_type
    except UnboundLocalError:
        raise ValueError(f'ID type detection failed for "{sid}".')


def get_folder(api, view):
    """Auxiliary function to get the cache folder belonging to an API,
    eventually create the folder.
    """
    from configparser import NoOptionError
    from pathlib import Path

    from pybliometrics.scopus.utils import CONFIG_FILE, DEFAULT_PATHS
    from pybliometrics.scopus.utils.create_config import create_config

    if not config.has_section('Directories'):
        create_config()
    try:
        parent = Path(config.get('Directories', api))
    except NoOptionError:
        parent = DEFAULT_PATHS[api]
        config.set('Directories', api, str(parent))
        with open(CONFIG_FILE, 'w') as ouf:
            config.write(ouf)
    try:
        folder = parent/view
    except TypeError:
        folder = parent
    folder.mkdir(parents=True, exist_ok=True)
    return folder



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for `scopus.CitationOverview` module."""

from collections import namedtuple
from nose.tools import assert_equal, assert_true

from pybliometrics.scopus import CitationOverview


co_eid = CitationOverview(["85068268027", "84930616647"],
                          refresh=30, start=2016, end=2020)
co_doi = CitationOverview(["10.1016/j.softx.2019.100263"],
                          id_type="doi", refresh=30, start=2016, end=2020)

def test_authors():
    Author = namedtuple('Author', 'name surname initials id url')
    url = 'https://api.elsevier.com/content/author/author_id/7004212771'
    john = Author(name='Kitchin J.R.', surname='Kitchin',
                  initials='J.R.', id='7004212771', url=url)
    assert_equal(co_eid.authors[0][1], john)
    assert_equal(co_eid.authors[1], [john])
    assert_equal(co_doi.authors[0][1], john)


def test_cc():
    expected0 = [(2016, 0), (2017, 0), (2018, 0), (2019, 0), (2020, 6)]
    expected1 = [(2016, 4), (2017, 2), (2018, 2), (2019, 2), (2020, 2)]
    assert_equal(co_eid.cc, [expected0, expected1])
    assert_equal(co_doi.cc, [expected0])


def test_citationType_long():
    assert_equal(co_eid.citationType_long, ['Article', 'Review'])
    assert_equal(co_doi.citationType_long, ['Article'])


def test_citationType_short():
    assert_equal(co_eid.citationType_short, ['ar', 're'])
    assert_equal(co_doi.citationType_short, ['ar'])


def test_columnTotal():
    assert_equal(co_eid.columnTotal, [4, 2, 2, 2, 8])
    assert_equal(co_doi.columnTotal, [0, 0, 0, 0, 6])


def test_doi():
    expected = ['10.1016/j.softx.2019.100263', '10.1021/acscatal.5b00538']
    assert_equal(co_eid.doi, expected)
    assert_equal(co_doi.doi, [expected[0]])


def test_endingPage():
    assert_equal(co_eid.endingPage, [None, '3899'])
    assert_equal(co_doi.endingPage, None)


def test_grandTotal():
    assert_true(co_eid.grandTotal >= 29)
    assert_true(co_doi.grandTotal >= 16)


def test_h_index():
    assert_equal(co_eid.h_index, 2)
    assert_equal(co_doi.h_index, 1)


def test_issn():
    expected = ['2352-7110', '2155-5435']
    assert_equal(co_eid.issn, expected)
    assert_equal(co_doi.issn, [expected[0]])


def test_issueIdentifier():
    assert_equal(co_eid.issueIdentifier, [None, '6'])
    assert_equal(co_doi.issueIdentifier, None)


def test_laterColumnTotal():
    assert_true(co_eid.laterColumnTotal >= 18)
    assert_true(co_doi.laterColumnTotal >= 16)


def test_lcc():
    assert_true(co_eid.lcc[0] >= 1)
    assert_true(co_eid.lcc[1] >= 1)
    assert_true(co_doi.lcc[0] >= 1)


def test_pcc():
    assert_true(co_eid.pcc == [0, 0])
    assert_true(co_doi.pcc == [0])


def test_pii():
    expected = ['S2352711019300573', None]
    assert_equal(co_eid.pii, expected)
    assert_equal(co_doi.pii, [expected[0]])


def prevColumnTotal():
    assert_equal(co_eid.prevColumnTotal, 0)
    assert_equal(co_doi.prevColumnTotal, 0)


def test_publicationName():
    expected = ['SoftwareX', 'ACS Catalysis']
    assert_equal(co_eid.publicationName, expected)
    assert_equal(co_doi.publicationName, [expected[0]])


def test_rangeColumnTotal():
    assert_equal(co_eid.rangeColumnTotal, 18)
    assert_equal(co_doi.rangeColumnTotal, 6)


def test_rangeCount():
    assert_true(co_eid.rangeCount[0] >= 6)
    assert_true(co_eid.rangeCount[1] >= 6)
    assert_true(co_doi.rangeCount[0] >= 6)


def test_rowTotal():
    assert_true(co_eid.rowTotal[0] >= 10)
    assert_true(co_eid.rowTotal[1] >= 10)
    assert_true(co_doi.rowTotal[0] >= 10)


def test_scopus_id():
    expected = [85068268027, 84930616647]
    assert_equal(co_eid.scopus_id, expected)
    assert_equal(co_doi.scopus_id, [expected[0]])


def test_startingPage():
    assert_equal(co_eid.startingPage, [None, '3894'])
    assert_equal(co_doi.startingPage, None)


def test_title():
    expected = ['pybliometrics: Scriptable bibliometrics using a Python interface to Scopus',
                'Examples of effective data sharing in scientific publishing']
    assert_equal(co_eid.title, expected)
    assert_equal(co_doi.title, [expected[0]])


def test_url():
    expected = ['https://api.elsevier.com/content/abstract/scopus_id/85068268027',
                'https://api.elsevier.com/content/abstract/scopus_id/84930616647']
    assert_equal(co_eid.url, expected)
    assert_equal(co_doi.url, [expected[0]])


def test_volume():
    assert_equal(co_eid.volume, ['10', '5'])
    assert_equal(co_doi.volume, ['10'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for `scopus.AuthorSearch` module."""

from collections import namedtuple
from nose.tools import assert_equal, assert_true

from pybliometrics.scopus import AuthorSearch

s1 = AuthorSearch('authlast(selten) and authfirst(reinhard)', refresh=30)
s2 = AuthorSearch('authlast(selten)', download=False, refresh=True)


def test_authors():
    assert_true(isinstance(s1.authors, list))
    assert_true(len(s1.authors) >= 2)
    order = 'eid orcid surname initials givenname affiliation documents '\
            'affiliation_id city country areas'
    Author = namedtuple('Author', order)
    expected = Author(eid='9-s2.0-6602907525', orcid=None, surname='Selten',
        initials='R.', givenname='Reinhard', affiliation='Universität Bonn',
        documents=74, affiliation_id='60007493', city='Bonn',
        country='Germany', areas='ECON (73); MATH (19); BUSI (16)')
    assert_equal(s1.authors[0], expected)


def test_authors_nodownload():
    # Only works if query hasn't been cached
    assert_equal(s2.authors, None)


def test_results_size():
    received1 = s1.get_results_size()
    assert_true(received1 >= 1)
    received2 = s2.get_results_size()
    assert_true(received2 >= 25)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for `scopus.PlumXMetrics` module."""

from nose.tools import assert_equal, assert_true

from pybliometrics.scopus import PlumXMetrics

# A published paper. All PlumX Metrics categories are present.
m1 = PlumXMetrics('2-s2.0-34249753618', 'elsevierId', refresh=30)
# A published paper. No Usage and Social Media metrics.
m2 = PlumXMetrics('10.2307/2281868', 'doi', refresh=30)
# A published paper. No PlumX Metrics.
m3 = PlumXMetrics('2-s2.0-84950369844', 'elsevierId', refresh=30)
# An arXiv paper. No Citation metrics.
m4 = PlumXMetrics('1507.02672', 'arxivId', refresh=30)
# A published book. No Social Media metrics.
m5 = PlumXMetrics('9783540783893', 'isbn', refresh=30)
# A GitHub repository. No Citation metrics.
m6 = PlumXMetrics('tensorflow/tensorflow', 'githubRepoId', refresh=30)
# A YouTube video. No Capture and Citation metrics.
m7 = PlumXMetrics('dQw4w9WgXcQ', 'youtubeVideoId', refresh=30)


def test_category_totals():
    m1_received = sorted([c.name  for c in m1.category_totals])
    m1_expected = ['capture', 'citation', 'mention', 'socialMedia', 'usage']
    assert_equal(m1_received, m1_expected)
    m2_received = sorted([c.name  for c in m2.category_totals])
    m2_expected = ['capture', 'citation', 'mention']
    assert_equal(m2_received, m2_expected)
    assert_equal(m3.category_totals, None)
    m4_received = sorted([c.name  for c in m4.category_totals])
    m4_expected = ['citation', 'mention', 'socialMedia']
    assert_equal(m4_received, m4_expected)
    m6_received = sorted([c.name  for c in m6.category_totals])
    m6_expected = ['capture', 'mention', 'socialMedia', 'usage']
    assert_equal(m6_received, m6_expected)
    m7_received = sorted([c.name  for c in m7.category_totals])
    m7_expected = ['mention', 'socialMedia', 'usage']
    assert_equal(m7_received, m7_expected)
    for plumx in (m1, m2, m4, m5, m6, m7):
        m_fields = set(field for ntup in plumx.category_totals for field in ntup._fields)
        assert_equal(m_fields, {'name', 'total'})
        zero_totals = [i.total for i in plumx.category_totals if i.total <= 0]
        assert_equal(zero_totals, [])


def test_capture():
    for plumx in (m3, m4, m7):
        assert_equal(plumx.capture, None)
    expected = {'name', 'total'}
    for plumx in (m1, m2, m5, m6):
        assert_true(isinstance(plumx.capture, list))
        assert_true(len(plumx.capture) > 0)
        m_fields = set(field for ntup in plumx.capture for field in ntup._fields)
        assert_equal(m_fields, expected)
        zero_totals = [i.total for i in plumx.capture if i.total <= 0] 
        assert_equal(zero_totals, [])


def test_citation():
    assert_equal(m6.citation, None)
    assert_equal(m7.citation, None)
    expected = {'name', 'total'}
    for plumx in (m1, m2, m4, m5):
        assert_true(isinstance(plumx.citation, list))
        assert_true(len(plumx.citation) > 0)
        m_fields = set(field for ntup in plumx.citation for field in ntup._fields)
        assert_equal(m_fields, expected)
        zero_totals = [i.total for i in plumx.citation if i.total <= 0]
        assert_equal(zero_totals, [])


def test_mention():
    assert_equal(m3.mention, None)
    expected = {'name', 'total'}
    for plumx in (m1, m2, m4, m5, m6, m7):
        assert_true(isinstance(plumx.mention, list))
        assert_true(len(plumx.mention) > 0)
        m_fields = set(field for ntup in plumx.mention for field in ntup._fields)
        assert_equal(m_fields, expected)
        zero_totals = [i.total for i in plumx.mention if i.total <= 0]
        assert_equal(zero_totals, [])


def test_social_media():
    assert_equal(m2.social_media, None)
    assert_equal(m3.social_media, None)
    assert_equal(m5.social_media, None)
    expected = {'name', 'total'}
    for plumx in (m1, m4, m6, m7):
        assert_true(isinstance(plumx.social_media, list))
        assert_true(len(plumx.social_media) > 0)
        m_fields = set(field for ntup in plumx.social_media for field in ntup._fields)
        assert_equal(m_fields, expected)
        zero_totals = [i.total for i in plumx.social_media if i.total <= 0]
        assert_equal(zero_totals, [])


def test_usage():
    assert_equal(m2.usage, None)
    assert_equal(m3.usage, None)
    assert_equal(m4.usage, None)
    expected = {'name', 'total'}
    for plumx in (m1, m5, m6, m7):
        assert_true(isinstance(plumx.usage, list))
        assert_true(len(plumx.usage) > 0)
        m_fields = set(field for ntup in plumx.usage for field in ntup._fields)
        assert_equal(m_fields, expected)
        zero_totals = [i.total for i in plumx.usage if i.total <= 0]
        assert_equal(zero_totals, [])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for `scopus.AuthorRetrieval` module."""

import warnings
from collections import Counter, namedtuple
from nose.tools import assert_equal, assert_true

from pybliometrics.scopus import AuthorRetrieval

warnings.simplefilter("always")

metrics = AuthorRetrieval("7004212771", refresh=30, view="METRICS")
light = AuthorRetrieval("7004212771", refresh=30, view="LIGHT")
standard = AuthorRetrieval("7004212771", refresh=30, view="STANDARD")
enhanced = AuthorRetrieval("7004212771", refresh=30, view="ENHANCED")


def test_affiliation_current():
    assert_equal(metrics.affiliation_current, None)
    assert_equal(light.affiliation_current, None)
    order = 'id parent type relationship afdispname preferred_name '\
            'parent_preferred_name country_code country address_part city '\
            'state postal_code org_domain org_URL'
    aff = namedtuple('Affiliation', order)
    expected = aff(id=110785688, parent=60027950, type='dept',
        relationship='author', afdispname=None, country='United States',
        preferred_name='Department of Chemical Engineering',
        parent_preferred_name='Carnegie Mellon University', country_code='usa',
        address_part='5000 Forbes Avenue', city='Pittsburgh', state='PA',
        postal_code='15213-3890', org_domain='cmu.edu',
        org_URL='https://www.cmu.edu/')
    for a in (standard, enhanced):
        received = a.affiliation_current
        assert_true(isinstance(received, list))
        assert_true(len(received) >= 1)
        assert_true(expected in received)


def test_affiliation_history():
    assert_equal(metrics.affiliation_history, None)
    assert_equal(light.affiliation_history, None)
    order = 'id parent type relationship afdispname preferred_name '\
            'parent_preferred_name country_code country address_part city '\
            'state postal_code org_domain org_URL'
    aff = namedtuple('Affiliation', order)
    expected = aff(id=60008644, parent=None, type='parent',
        relationship='author', afdispname=None,
        preferred_name='Fritz Haber Institute of the Max Planck Society',
        parent_preferred_name=None, country_code='deu', country='Germany',
        address_part='Faradayweg 4-6', city='Berlin', state=None,
        postal_code='14195', org_domain='fhi.mpg.de',
        org_URL='https://www.fhi.mpg.de/')
    for a in (standard, enhanced):
        received = a.affiliation_history
        assert_true(isinstance(received, list))
        assert_true(len(received) >= 10)
        assert_true(expected in received)


def test_alias():
    assert_equal(metrics.alias, None)
    assert_equal(light.alias, None)
    assert_equal(standard.alias, None)
    assert_equal(enhanced.alias, None)


def test_citation_count():
    expected = 13600
    assert_true(metrics.citation_count >= expected)
    assert_true(light.citation_count >= expected)
    assert_true(standard.citation_count >= expected)
    assert_true(enhanced.citation_count >= expected)


def test_cited_by_count():
    expected = 10900
    assert_true(metrics.cited_by_count >= expected)
    assert_true(light.cited_by_count >= expected)
    assert_true(standard.cited_by_count >= expected)
    assert_true(enhanced.cited_by_count >= expected)


def test_classificationgroup():
    assert_equal(metrics.classificationgroup, None)
    assert_equal(light.classificationgroup, None)
    for a in (standard, enhanced):
        received = a.classificationgroup
        assert_true(isinstance(received, list))
        assert_true(len(received) > 0)
        assert_true((1906, 1) in received)  # frequency might differ


def test_coauthor_count():
    assert_equal(light.coauthor_count, None)
    assert_equal(standard.coauthor_count, None)
    expected = 175
    assert_true(metrics.coauthor_count >= expected)
    assert_true(enhanced.coauthor_count >= expected)


def test_coauthor_link():
    assert_equal(metrics.coauthor_link, None)
    assert_equal(light.coauthor_link, None)
    assert_equal(standard.coauthor_link, None)
    expected = 'http://api.elsevier.com/content/search/author?co-author=7004212771'
    assert_equal(enhanced.coauthor_link, expected)


def test_date_created():
    assert_equal(metrics.date_created, None)
    assert_equal(light.date_created, None)
    assert_equal(standard.date_created, (2005, 12, 3))
    assert_equal(enhanced.date_created, (2005, 12, 3))


def test_document_count():
    expected = 106
    assert_true(light.document_count >= expected)
    assert_true(standard.document_count >= expected)
    assert_true(metrics.document_count >= expected)
    assert_true(enhanced.document_count >= expected)


def test_eid():
    assert_equal(metrics.eid, None)
    assert_equal(light.eid, '9-s2.0-7004212771')
    assert_equal(standard.eid, '9-s2.0-7004212771')
    assert_equal(enhanced.eid, '9-s2.0-7004212771')


def test_estimate_uniqueness():
    expected = 2
    assert_equal(metrics.estimate_uniqueness(), 0)
    assert_equal(light.estimate_uniqueness(), 0)
    assert_equal(standard.estimate_uniqueness(), expected)
    assert_equal(enhanced.estimate_uniqueness(), expected)


def test_given_name():
    assert_equal(metrics.given_name, None)
    assert_equal(light.given_name, None)
    assert_equal(standard.given_name, 'John R.')
    assert_equal(enhanced.given_name, 'John R.')


def get_coauthors():
    assert_equal(metrics.get_coauthors(), None)
    assert_equal(light.get_coauthors(), None)
    assert_equal(standard.get_coauthors(), None)
    received = enhanced.get_coauthors()
    assert_true(isinstance(received, list))
    assert_true(len(received) > 155)
    fields = 'surname given_name id areas affiliation_id name city country'
    coauth = namedtuple('Coauthor', fields)
    expected = coauth(surname='Rose', given_name='Michael E.', id=57209617104,
        areas='Computer Science (all)', affiliation_id=60105007,
        name='Max-Planck-Institut für Innovation und Wettbewerb',
        city='Munich', country='Germany')
    assert_true(expected in received)


def test_get_documents():
    subtypes = {'re', 'ed', 'no'}
    received = enhanced.get_documents(subtypes)
    assert_equal(len(received), 7)


def test_get_document_eids():
    expected = 100
    assert_true(len(enhanced.get_document_eids()) >= expected)


def test_h_index():
    assert_equal(light.h_index, None)
    assert_equal(standard.h_index, None)
    expected = 34
    assert_true(metrics.h_index >= expected)
    assert_true(enhanced.h_index >= expected)


def test_historical_identifier():
    assert_equal(metrics.historical_identifier, None)
    assert_equal(light.historical_identifier, None)
    assert_equal(standard.historical_identifier, None)
    expected = [35787230500, 36488127000, 54974425600, 55004143700,
                55004143800, 57057263700, 56641032000, 36747787600,
                57206217299, 57219840256]
    assert_equal(enhanced.historical_identifier, expected)


def test_identifier():
    expected = 7004212771
    assert_equal(metrics.identifier, expected)
    assert_equal(light.identifier, expected)
    assert_equal(standard.identifier, expected)
    assert_equal(enhanced.identifier, expected)


def test_indexed_name():
    assert_equal(metrics.indexed_name, None)
    assert_equal(light.indexed_name, None)
    assert_equal(standard.indexed_name, 'Kitchin J.')
    assert_equal(enhanced.indexed_name, 'Kitchin J.')


def test_initials():
    assert_equal(metrics.initials, None)
    assert_equal(light.initials, None)
    assert_equal(standard.initials, 'J.R.')
    assert_equal(enhanced.initials, 'J.R.')


def test_name_variants():
    assert_equal(metrics.name_variants, None)
    assert_equal(light.name_variants, None)
    expected = "<class 'pybliometrics.scopus.author_retrieval.Variant'>"
    for a in (standard, enhanced):
        received = a.name_variants
        assert_true(isinstance(received, list))
        assert_true(len(received) > 0)
        assert_equal(str(type(received[0])), expected)


def test_orcid():
    assert_equal(metrics.orcid, None)
    assert_equal(light.orcid, '0000-0003-2625-9232')
    assert_equal(standard.orcid, '0000-0003-2625-9232')
    assert_equal(enhanced.orcid, '0000-0003-2625-9232')


def test_publication_range():
    assert_equal(metrics.publication_range, None)
    assert_equal(light.publication_range, None)
    for a in (standard, enhanced):
        assert_equal(a.publication_range[0], 1995)
        assert_true(a.publication_range[1] >= 2021)


def test_scopus_author_link():
    assert_equal(metrics.scopus_author_link, None)
    expected = 'http://api.elsevier.com/content/author/author_id/7004212771'
    assert_equal(light.scopus_author_link, expected)
    assert_equal(standard.scopus_author_link, expected)
    assert_equal(enhanced.scopus_author_link, expected)


def test_search_link():
    assert_equal(metrics.search_link, None)
    expected = 'http://api.elsevier.com/content/search/scopus?query='\
               'au-id%287004212771%29'
    assert_equal(light.search_link, expected)
    assert_equal(standard.search_link, expected)
    assert_equal(enhanced.search_link, expected)


def test_self_link():
    assert_equal(metrics.self_link, None)
    expected = 'https://www.scopus.com/authid/detail.uri?partnerID=HzOxMe3b&'\
               'authorId=7004212771&origin=inward'
    assert_equal(light.self_link, expected)
    assert_equal(standard.self_link, expected)
    assert_equal(enhanced.self_link, expected)


def test_status():
    assert_equal(metrics.status, None)
    assert_equal(light.status, None)
    expected = "update"
    assert_equal(standard.status, expected)
    assert_equal(enhanced.status, expected)


def test_subject_areas():
    assert_equal(metrics.subject_areas, None)
    assert_equal(light.subject_areas, None)
    expected = "<class 'pybliometrics.scopus.author_retrieval.Subjectarea'>"
    for a in (standard, enhanced):
        received = a.subject_areas
        assert_true(isinstance(received, list))
        assert_true(len(received) > 0)
        assert_equal(str(type(received[0])), expected)


def test_surname():
    assert_equal(metrics.surname, None)
    assert_equal(light.surname, None)
    assert_equal(standard.surname, 'Kitchin')
    assert_equal(enhanced.surname, 'Kitchin')


def test_url():
    expected = 'http://api.elsevier.com/content/author/author_id/7004212771'
    assert_equal(metrics.url, expected)
    assert_equal(light.url, expected)
    assert_equal(standard.url, expected)
    assert_equal(enhanced.url, expected)


def test_warning_without_forwarding():
    with warnings.catch_warnings(record=True) as w:
        au = AuthorRetrieval("24079538400", refresh=False)
        assert_equal(len(w), 1)
        assert_true(issubclass(w[-1].category, UserWarning))
        assert_true("24079538400" in str(w[-1].message))


def test_warning_with_forwarding():
    au = AuthorRetrieval("57191449583", refresh=False)
    with warnings.catch_warnings(record=True) as w:
        auth_id = au.identifier
        assert_equal(len(w), 1)
        assert_true(issubclass(w[-1].category, UserWarning))
        assert_true("57191449583" in str(w[-1].message))
    assert_equal(auth_id, 36854449200)



from collections import namedtuple
from warnings import warn


def chained_get(container, path, default=None):
    """Helper function to perform a series of .get() methods on a dictionary
    or return the `default`.

    Parameters
    ----------
    container : dict
        The dictionary on which the .get() methods should be performed.

    path : list or tuple
        The list of keys that should be searched for.

    default : any (optional, default=None)
        The object type that should be returned if the search yields
        no result.
    """
    from functools import reduce

    # Obtain value via reduce
    try:
        return reduce(lambda c, k: c.get(k, default), path, container)
    except (AttributeError, TypeError):
        return default


def check_integrity(tuples, fields, action):
    """Check integrity of specific fields in a list of tuples and perfom
    provided action.
    """
    for field in fields:
        elements = [getattr(e, field) for e in tuples]
        if None not in elements:
            continue
        msg = "Parsed information doesn't pass integrity check because of "\
              f"incomplete information in field '{field}'"
        if action == "raise":
            raise AttributeError(msg)
        elif action == "warn":
            warn(msg)


def check_field_consistency(needles, haystack):
    """Raise ValueError if elements of a list are not present in a string."""
    wrong = set(needles) - set(haystack.split())
    if wrong:
        msg = f"Element(s) '{', '.join(sorted(wrong))}' not allowed in "\
              "parameter integrity_fields"
        raise ValueError(msg)


def get_id(s, integer=True):
    """Helper function to return the Scopus ID at a fixed position."""
    path = ['coredata', 'dc:identifier']
    try:
        return int(chained_get(s, path, "").split(':')[-1])
    except ValueError:
        return None


def get_link(dct, idx, path=['coredata', 'link']):
    """Helper function to return the link at position `idx` from coredata."""
    links = chained_get(dct, path, [{}])
    try:
        return links[idx].get('@href')
    except IndexError:
        return None


def listify(element):
    """Helper function to turn an element into a list if it isn't a list yet.
    """
    if isinstance(element, list):
        return element
    else:
        return [element]


def make_int_if_possible(val):
    """Attempt a conversion to int type."""
    try:
        return int(val)
    except TypeError:
        return val


def make_search_summary(self, keyword, results, joiner="\n    "):
    """Create string for str dunder of search classes."""
    date = self.get_cache_file_mdate().split()[0]
    if self._n != 1:
        appendix = "s"
        verb = "have"
    else:
        appendix = ""
        verb = "has"
    s = f"Search '{self._query}' yielded {self._n:,} "\
        f"{keyword}{appendix} as of {date}"
    if results:
        s += ":" + joiner + joiner.join(results)
    elif self._n:
        s += f", which {verb} not been downloaded"
    return s


def parse_affiliation(affs):
    """Helper function to parse list of affiliation-related information."""
    order = 'id parent type relationship afdispname preferred_name '\
            'parent_preferred_name country_code country address_part city '\
            'state postal_code org_domain org_URL'
    aff = namedtuple('Affiliation', order)
    out = []
    for item in listify(affs):
        if not item:
            continue
        doc = item.get('ip-doc', {}) or {}
        address = doc.get('address', {}) or {}
        try:
            parent = int(item['@parent'])
        except KeyError:
            parent = None
        new = aff(id=int(item['@affiliation-id']), parent=parent,
            type=doc.get('@type'), relationship=doc.get('@relationship'),
            afdispname=doc.get('@afdispname'),
            preferred_name=doc.get('preferred-name', {}).get('$'),
            parent_preferred_name=doc.get('parent-preferred-name', {}).get('$'),
            country_code=address.get('@country'), country=address.get('country'),
            address_part=address.get("address-part"), city=address.get('city'),
            postal_code=address.get('postal-code'), state=address.get('state'),
            org_domain=doc.get('org-domain'), org_URL=doc.get('org-URL'))
        out.append(new)
    return out or None


def parse_date_created(dct):
    """Helper function to parse date-created from profile."""
    date = dct['date-created']
    if date:
        return int(date['@year']), int(date['@month']), int(date['@day'])
    else:
        return None, None, None



from collections import namedtuple
from typing import List, NamedTuple, Optional, Tuple, Union

from pybliometrics.scopus.superclasses import Search
from pybliometrics.scopus.utils import check_integrity, check_parameter_value,\
    check_field_consistency, listify, make_search_summary


class AuthorSearch(Search):
    @property
    def authors(self) -> Optional[List[NamedTuple]]:
        """A list of namedtuples storing author information,
        where each namedtuple corresponds to one author.
        The information in each namedtuple is (eid orcid surname initials givenname
        documents affiliation affiliation_id city country areas).

        All entries are strings or None.  Areas combines abbreviated subject
        areas followed by the number of documents in this subject.

        Raises
        ------
        ValueError
            If the elements provided in integrity_fields do not match the
            actual field names (listed above).
        """
        # Initiate namedtuple with ordered list of fields
        fields = 'eid orcid surname initials givenname affiliation documents '\
                 'affiliation_id city country areas'
        auth = namedtuple('Author', fields)
        check_field_consistency(self._integrity, fields)
        # Parse elements one-by-one
        out = []
        for item in self._json:
            name = item.get('preferred-name', {})
            aff = item.get('affiliation-current', {})
            fields = item.get('subject-area',
                              [{'@abbrev': '', '@frequency': ''}])
            areas = [f"{d.get('@abbrev', '')} ({d.get('@frequency', '')})"
                     for d in listify(fields)]
            new = auth(eid=item.get('eid'),
                       orcid=item.get('orcid'),
                       initials=name.get('initials'),
                       surname=name.get('surname'),
                       areas="; ".join(areas),
                       givenname=name.get('given-name'),
                       documents=int(item['document-count']),
                       affiliation=aff.get('affiliation-name'),
                       affiliation_id=aff.get('affiliation-id'),
                       city=aff.get('affiliation-city'),
                       country=aff.get('affiliation-country'))
            out.append(new)
        # Finalize
        check_integrity(out, self._integrity, self._action)
        return out or None

    def __init__(self,
                 query: str,
                 refresh: Union[bool, int] = False,
                 verbose: bool = False,
                 download: bool = True,
                 integrity_fields: Union[List[str], Tuple[str, ...]] = None,
                 integrity_action: str = "raise",
                 count: int = 200,
                 **kwds
                 ) -> None:
        """Interaction with the Author Search API.

        :param query: A string of the query.  For allowed fields and values see
                      https://dev.elsevier.com/sc_author_search_tips.html.
        :param refresh: Whether to refresh the cached file if it exists or not.
                        If int is passed, cached file will be refreshed if the
                        number of days since last modification exceeds that value.
        :param download: Whether to download results (if they have not been
                         cached).
        :param verbose: Whether to print a download progress bar.
        :param integrity_fields: Names of fields whose completeness should
                                 be checked.  ScopusSearch will perform the
                                 action specified in `integrity_action` if
                                 elements in these fields are missing.  This
                                 helps avoiding idiosynchratically missing
                                 elements that should always be present
                                 (e.g., EID or source ID).
        :param integrity_action: What to do in case integrity of provided fields
                                 cannot be verified.  Possible actions:
                                 - "raise": Raise an AttributeError
                                 - "warn": Raise a UserWarning
        :param count: (deprecated) The number of entries to be displayed at
                      once.  A smaller number means more queries with each
                      query having fewer results.
        :param kwds: Keywords passed on as query parameters.  Must contain
                     fields and values mentioned in the API specification at
                     https://dev.elsevier.com/documentation/AuthorSearchAPI.wadl.

        Raises
        ------
        ScopusQueryError
            If the number of search results exceeds 5000, which is the API's
            maximum number of results returned.  The error prevents the
            download attempt and avoids making use of your API key.

        ValueError
            If any of the parameters `integrity_action` or `refresh` is not
            one of the allowed values.

        Notes
        -----
        The directory for cached results is `{path}/STANDARD/{fname}`,
        where  `path` is specified in your configuration file, and `fname` is
        the md5-hashed version of `query`.
        """
        # Checks
        allowed = ("warn", "raise")
        check_parameter_value(integrity_action, allowed, "integrity_action")
        if count != 200:
            msg = "Parameter `count` is deprecated and will be removed in a "\
                  "future release.  There will be no substitute."
            warn(msg, FutureWarning)

        # Query
        self._action = integrity_action
        self._integrity = integrity_fields or []
        self._query = query
        self._refresh = refresh
        self._view = "STANDARD"
        Search.__init__(self, query=query, api='AuthorSearch', count=count,
                        download=download, verbose=verbose, **kwds)

    def __str__(self):
        """Print a summary string."""
        names = [f'{n["preferred-name"].get("surname", "?")}, '
                 f'{n["preferred-name"].get("given-name", "?")}; '
                 f'{n["dc:identifier"]} ({int(n["document-count"]):,} document(s))'
                 for n in self._json]
        return make_search_summary(self, "author", names)



from pbr.version import VersionInfo

_v = VersionInfo('pybliometrics').semantic_version()
__version__ = _v.release_string()
version_info = _v.version_tuple()

__citation__ = 'Rose, Michael E. and John R. Kitchin: "pybliometrics: '\
    'Scriptable bibliometrics using a Python interface to Scopus", SoftwareX '\
    '10 (2019) 100263.'

import pybliometrics.scopus



"""Base class object for superclasses."""

from json import dumps, loads
from time import localtime, strftime, time
from typing import Dict, Optional

from pybliometrics.scopus.exception import ScopusQueryError
from pybliometrics.scopus.utils import get_content, SEARCH_MAX_ENTRIES
from tqdm import tqdm


class Base:
    def __init__(self,
                 params: Dict,
                 url: str,
                 api: str,
                 download: bool = True,
                 verbose: bool = False,
                 *args: str, **kwds: str
                 ) -> None:
        """Class intended as base class for superclasses.

        :param params: Dictionary used as header during the API request.
        :param url: The URL to be accessed.
        :param api: The Scopus API to be accessed.
        :param download: Whether to download the query or not.  Has no effect
                         for retrieval requests.
        :param verbose: Whether to print a download progress bar.
        :param args: Keywords passed on `get_content()`
        :param kwds: Keywords passed on `get_content()`

        Raises
        ------
        ValueError
            If `self._refresh` is neither boolean nor numeric.
        """
        # Checks
        try:
            _ = int(self._refresh)
        except ValueError:
            msg = "Parameter refresh needs to be numeric or boolean."
            raise ValueError(msg)

        # Compare age of file to test whether we refresh
        self._refresh, mod_ts = _check_file_age(self)

        # Read or download, possibly with caching
        fname = self._cache_file_path
        search_request = "query" in params
        if fname.exists() and not self._refresh:
            self._mdate = mod_ts
            if search_request:
                self._json = [loads(line) for line in
                              fname.read_text().split("\n") if line]
                self._n = len(self._json)
            else:
                self._json = loads(fname.read_text())
        else:
            resp = get_content(url, api, params, *args, **kwds)
            header = resp.headers
            if search_request:
                # Get number of results
                res = resp.json()
                n = int(res['search-results'].get('opensearch:totalResults', 0))
                self._n = n
                self._json = []
                # Results size check
                if params.get("cursor") is None and n > SEARCH_MAX_ENTRIES:
                    # Stop if there are too many results
                    text = f'Found {n} matches.  The query fails to return '\
                           f'more than {SEARCH_MAX_ENTRIES} entries.  Change '\
                           'your query such that it returns fewer entries.'
                    raise ScopusQueryError(text)
                # Download results page-wise
                if download:
                    data = ""
                    if n:
                        data, header = _parse(res, n, url, api, params,
                                              verbose, *args, **kwds)
                        self._json = data
                else:
                    data = None
            else:
                data = loads(resp.text)
                self._json = data
                data = [data]
            # Set private variables
            self._mdate = time()
            self._header = header
            # Finally write data unless download=False
            if download:
                text = [dumps(item, separators=(',', ':')) for item in data]
                fname.write_text("\n".join(text))

    def get_cache_file_age(self) -> int:
        """Return the age of the cached file in days."""
        diff = time() - self._mdate
        return int(diff / 86400)

    def get_cache_file_mdate(self) -> str:
        """Return the modification date of the cached file."""
        return strftime('%Y-%m-%d %H:%M:%S', localtime(self._mdate))

    def get_key_remaining_quota(self) -> Optional[str]:
        """Return number of remaining requests for the current key and the
        current API (relative on last actual request).
        """
        try:
            return self._header['X-RateLimit-Remaining']
        except AttributeError:
            return None

    def get_key_reset_time(self) -> Optional[str]:
        """Return time when current key is reset (relative on last
        actual request).
        """
        try:
            date = int(self._header['X-RateLimit-Reset'])
            return strftime('%Y-%m-%d %H:%M:%S', localtime(date))
        except AttributeError:
            return None


def _check_file_age(self):
    """Whether a file needs to be refreshed based on its age."""
    refresh = self._refresh
    try:
        mod_ts = self._cache_file_path.stat().st_mtime
        if not isinstance(refresh, bool):
            diff = time() - mod_ts
            days = int(diff / 86400) + 1
            allowed_age = int(self._refresh)
            refresh = allowed_age < days
    except FileNotFoundError:
        refresh = True
        mod_ts = None
    return refresh, mod_ts


def _parse(res, n, url, api, params, verbose, *args, **kwds):
    """Auxiliary function to download results and parse json."""
    cursor = "cursor" in params
    if not cursor:
        start = params["start"]
    _json = res.get('search-results', {}).get('entry', [])
    if verbose:
        # Roundup + 1 for the final iteration
        print(f'Downloading results for query "{params["query"]}":')
        n_chunks = int(n/params['count']) + (n % params['count'] > 0) + 1
        pbar = tqdm(total=n_chunks)
        pbar.update(1)
    # Download the remaining information in chunks
    while n > 0:
        n -= params["count"]
        if cursor:
            pointer = res['search-results']['cursor'].get('@next')
            params.update({'cursor': pointer})
        else:
            start += params["count"]
            params.update({'start': start})
        resp = get_content(url, api, params, *args, **kwds)
        res = resp.json()
        _json.extend(res.get('search-results', {}).get('entry', []))
        if verbose:
            pbar.update(1)
    if verbose:
        pbar.close()
    return _json, resp.headers



import configparser
from typing import List, Optional


def create_config(keys: Optional[List[str]] = None,
                  insttoken: Optional[str] = None
                  ):
    """Initiates process to generate configuration file.

    :param keys: If you provide a list of keys, pybliometrics will skip the
                 prompt.  It will also not ask for InstToken.  This is
                 intended for workflows using CI, not for general use.
    :param insttoken: An InstToken to be used alongside the key(s).  Will only
                      be used if `keys` is not empty.
    """
    from pybliometrics.scopus.utils.constants import CONFIG_FILE, DEFAULT_PATHS

    config = configparser.ConfigParser()
    config.optionxform = str
    print(f"Creating config file at {CONFIG_FILE} with default paths...")
    # Set directories
    config.add_section('Directories')
    for api, path in DEFAULT_PATHS.items():
        config.set('Directories', api, str(path))

    # Set authentication
    config.add_section('Authentication')
    if keys:
        if not isinstance(keys, list):
            raise ValueError("Parameter `keys` must be a list.")
        key = ", ".join(keys)
        token = insttoken
    else:
        prompt_key = "Please enter your API Key(s), obtained from "\
                     "http://dev.elsevier.com/myapikey.html.  Separate "\
                     "multiple keys by comma:\n"
        key = input(prompt_key)
        prompt_token = "API Keys are sufficient for most users.  If you "\
                       "have an InstToken, please enter the token now; "\
                       "otherwise just press Enter:\n"
        token = input(prompt_token)
    config.set('Authentication', 'APIKey', key)
    if token:
        config.set('Authentication', 'InstToken', token)

    # Write out
    CONFIG_FILE.parent.mkdir(parents=True, exist_ok=True)
    with open(CONFIG_FILE, "w") as ouf:
        config.write(ouf)
    print(f"Configuration file successfully created at {CONFIG_FILE}\n"
          "For details see https://pybliometrics.rtfd.io/en/stable/configuration.html.")
    return config



import configparser
from collections import deque
from pathlib import Path

from pybliometrics.scopus.utils.constants import CONFIG_FILE, RATELIMITS
from pybliometrics.scopus.utils.create_config import create_config

# Read/create config file (with fixture for RTFD.io)
config = configparser.ConfigParser()
config.optionxform = str
try:
    if not CONFIG_FILE.exists():
        config = create_config()
    else:
        config.read(CONFIG_FILE)
    KEYS = [k.strip() for k in config.get('Authentication', 'APIKey').split(",")]
except EOFError:
    pass

# Throttling params
_throttling_params = {k: deque(maxlen=v) for k, v in RATELIMITS.items()}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for `scopus.SubjectClassifications` module."""

from nose.tools import assert_equal, assert_true

from pybliometrics.scopus import SubjectClassifications


# Search by words in subject description
sub1 = SubjectClassifications({'description': 'Physics'}, refresh=30)
# Search by subject code
sub2 = SubjectClassifications({'code': '2613'}, refresh=30)
# Search by words in subject detail
sub3 = SubjectClassifications({'detail': 'Processes'}, refresh=30)
# Search by subject abbreviation
sub4 = SubjectClassifications({'abbrev': 'MATH'}, refresh=30)
# Search by multiple criteria
sub5 = SubjectClassifications({'description': 'Engineering', 'detail': 'Fluid'}, refresh=30)
# Search by multiple criteria, subset returned fields
sub6 = SubjectClassifications({'detail': 'Analysis', 'description': 'Mathematics'},
                              fields=['description', 'detail'], refresh=30)


def test_results_desc():
    assert_true(len(sub1.results) > 0)
    assert_true(all(['Physics' in res.description for res in sub1.results]))


def test_results_code():
    assert_equal(len(sub2.results), 1)
    assert_equal(sub2.results[0].code, '2613')
    

def test_results_detail():
    assert_true(len(sub3.results) > 0)
    assert_true(all(['Processes' in res.detail for res in sub3.results]))


def test_results_abbrev():
    assert_true(len(sub4.results) > 0)
    assert_true(all(['MATH' in res.abbrev for res in sub4.results]))


def test_results_multi():
    assert_true(len(sub5.results) > 0)
    assert_true(all(['Engineering' in res.description for res in sub5.results]))
    assert_true(all(['Fluid' in res.detail for res in sub5.results]))
    

def test_results_fields():
    assert_true(len(sub6.results) > 0)
    assert_true(all(['Mathematics' in res.description for res in sub6.results]))
    assert_true(all(['Analysis' in res.detail for res in sub6.results]))
    assert_true(all([set(res._fields) == set(['description', 'detail']) for res in sub6.results]))



from collections import namedtuple
from typing import List, NamedTuple, Optional, Union

from pybliometrics.scopus.superclasses import Retrieval
from pybliometrics.scopus.utils import check_parameter_value


class PlumXMetrics(Retrieval):
    @property
    def category_totals(self) -> Optional[List[NamedTuple]]:
        """A list of namedtuples representing total metrics as categorized
        by PlumX Metrics in the form (capture, citation, mention, socialMedia,
        usage).

        Note: For Citation category a maximum citation count across sources
        is shown.  For details on PlumX Metrics categories see
        https://plumanalytics.com/learn/about-metrics/.
        """
        categories = self._json.get('count_categories', [])
        return _format_as_namedtuple_list(categories, "Category") or None

    @property
    def capture(self) -> Optional[List[NamedTuple]]:
        """A list of namedtuples representing metrics in the Captures category.

        Note: For details on Capture metrics see
        https://plumanalytics.com/learn/about-metrics/capture-metrics/.
        """
        metrics = self._count_categories.get('capture', [])
        return _format_as_namedtuple_list(metrics) or None

    @property
    def citation(self) -> Optional[List[NamedTuple]]:
        """A list of namedtuples representing citation counts from
        different sources.

        Note: For details on Citation metrics see
        https://plumanalytics.com/learn/about-metrics/citation-metrics/.
        """
        metrics = []
        for item in self._count_categories.get('citation', []):
            if item.get('sources'):
                metrics += item['sources']
        return _format_as_namedtuple_list(metrics) or None

    @property
    def mention(self) -> Optional[List[NamedTuple]]:
        """A list of namedtuples representing metrics in Mentions category.

        Note: For details on Mention metrics see
        https://plumanalytics.com/learn/about-metrics/mention-metrics/.
        """
        metrics = self._count_categories.get('mention', [])
        return _format_as_namedtuple_list(metrics) or None

    @property
    def social_media(self) -> Optional[List[NamedTuple]]:
        """A list of namedtuples representing social media metrics.

        Note: For details on Social Media metrics see
        https://plumanalytics.com/learn/about-metrics/social-media-metrics/.
        """
        metrics = self._count_categories.get('socialMedia', [])
        return _format_as_namedtuple_list(metrics) or None

    @property
    def usage(self) -> Optional[List[NamedTuple]]:
        """A list of namedtuples representing Usage category metrics.

        Note: For details on Usage metrics see
        https://plumanalytics.com/learn/about-metrics/usage-metrics/.
        """
        metrics = self._count_categories.get('usage', [])
        return _format_as_namedtuple_list(metrics) or None

    def __init__(self,
                 identifier: str,
                 id_type: str,
                 refresh: Union[bool, int] = False
                 ) -> None:
        """Interaction with the PlumX Metrics API.

        :param identifier: The identifier of a document.
        :param id_type: The type of used ID. Allowed values are:
                        - 'airitiDocId'
                        - 'arxivId'
                        - 'cabiAbstractId'
                        - 'citeulikeId'
                        - 'digitalMeasuresArtifactId'
                        - 'doi'
                        - 'elsevierId'
                        - 'elsevierPii'
                        - 'facebookCountUrlId'
                        - 'figshareArticleId'
                        - 'githubRepoId'
                        - 'isbn'
                        - 'lccn'
                        - 'medwaveId'
                        - 'nctId'
                        - 'oclc'
                        - 'pittEprintDscholarId'
                        - 'pmcid'
                        - 'pmid'
                        - 'redditId'
                        - 'repecHandle'
                        - 'repoUrl'
                        - 'scieloId'
                        - 'sdEid'
                        - 'slideshareUrlId'
                        - 'smithsonianPddrId'
                        - 'soundcloudTrackId'
                        - 'ssrnId'
                        - 'urlId'
                        - 'usPatentApplicationId'
                        - 'usPatentPublicationId'
                        - 'vimeoVideoId'
                        - 'youtubeVideoId'
        :param refresh: Whether to refresh the cached file if it exists or not.
                        If int is passed, cached file will be refreshed if the
                        number of days since last modification exceeds that value.

        Raises
        ------
        ValueError
            If the parameter `refresh` is not one of the allowed values.

        Notes
        -----
        The directory for cached results is `{path}/ENHANCED/{identifier}`,
        where `path` is specified in your configuration file.
        """
        # Checks
        allowed = ('airitiDocId', 'arxivId', 'cabiAbstractId',
                   'citeulikeId', 'digitalMeasuresArtifactId', 'doi',
                   'elsevierId', 'elsevierPii', 'facebookCountUrlId',
                   'figshareArticleId', 'githubRepoId', 'isbn',
                   'lccn', 'medwaveId', 'nctId', 'oclc',
                   'pittEprintDscholarId', 'pmcid', 'pmid', 'redditId',
                   'repecHandle', 'repoUrl', 'scieloId', 'sdEid',
                   'slideshareUrlId', 'smithsonianPddrId', 'soundcloudTrackId',
                   'ssrnId', 'urlId', 'usPatentApplicationId',
                   'usPatentPublicationId', 'vimeoVideoId', 'youtubeVideoId')
        check_parameter_value(id_type, allowed, "id_type")
        self._id_type = id_type
        self._identifier = identifier

        # Load json
        self._refresh = refresh
        self._view = 'ENHANCED'
        Retrieval.__init__(self, identifier=identifier, id_type=id_type,
                           api='PlumXMetrics')
        cats = self._json.get('count_categories', [])
        self._count_categories = {d["name"]: d['count_types'] for d in cats}

    def __str__(self):
        """Print a summary string."""
        s = f"Document with {self._id_type} {self._identifier} received:\n- "
        cats = [f"{c.total:,} citation(s) in category '{c.name}'"
                for c in self.category_totals]
        s += "\n- ".join(cats)
        s += f"\nas of {self.get_cache_file_mdate().split()[0]}"
        return s


def _format_as_namedtuple_list(metric_counts, tuple_name='Metric'):
    """Formats list of dicts of metrics into list of namedtuples in the
    form (name, total).
    """
    metric = namedtuple(tuple_name, 'name total')
    return [metric(name=t['name'], total=t['total']) for t in metric_counts]



"""Superclass to access all Scopus search APIs and dump the results."""

from hashlib import md5

from pybliometrics.scopus.superclasses import Base
from pybliometrics.scopus.utils import get_folder, URLS


class Search(Base):
    def __init__(self,
                 query: str,
                 api: str,
                 count: int = 200,
                 cursor: bool = False,
                 download: bool = True,
                 verbose: bool = False,
                 **kwds: str
                 ) -> None:
        """Class intended as superclass to perform a search query.

        :param query : A string of the query.
        :param api: The name of the Scopus API to be accessed.  Allowed values:
                    AffiliationSearch, AuthorSearch, ScopusSearch,
                    SerialSearch, SubjectClass.
        :param count: The number of entries to be displayed at once.  A smaller
                      number means more queries with each query having
                      fewer results.
        :param cursor: Whether to use the cursor in order to iterate over all
                      search results without limit on the number of the results.
                      In contrast to `start` parameter, the `cursor` parameter
                      does not allow users to obtain partial results.
        :param download: Whether to download results (if they have not been
                         cached) or not.
        :param verbose: Whether to print a download progress bar.
        :param kwds: Keywords passed on to requests header.  Must contain
                     fields and values specified in the respective API specification.

        Raises
        ------
        ValueError
            If the api parameter is an invalid entry.
        """
        # Construct query parameters
        params = {'count': count, 'view': self._view, **kwds}
        if isinstance(query, dict):
            params.update(query)
            name = "&".join(["=".join(t) for t in zip(query.keys(), query.values())])
        else:
            params['query'] = query
            name = query
        if cursor:
            params.update({'cursor': '*'})
        else:
            params.update({'start': 0})

        # Construct cache file path
        stem = md5(name.encode('utf8')).hexdigest()
        self._cache_file_path = get_folder(api, self._view)/stem

        # Init
        Base.__init__(self, params=params, url=URLS[api], download=download,
                      api=api, verbose=verbose)

    def get_results_size(self) -> int:
        """Return the number of results (works even if download=False)."""
        return self._n



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for `scopus.SerialTitle` module."""

from collections import namedtuple
from nose.tools import assert_equal, assert_true

from pybliometrics.scopus import SerialTitle

# SoftwareX
sofwarex = SerialTitle("2352-7110", refresh=30, years="2019-2020")
# OECD Economic Studies
oecd = SerialTitle("0255-0822", refresh=30)


def test_aggregation_type():
    assert_equal(sofwarex.aggregation_type, "journal")
    assert_equal(oecd.aggregation_type, "journal")


def test_citescoreyearinfolist():
    expected1 = [(2020, 2.8), (2021, 3.8)]
    assert_equal(sofwarex.citescoreyearinfolist, expected1)
    assert_equal(oecd.citescoreyearinfolist, None)


def test_eissn():
    assert_equal(sofwarex.eissn, "2352-7110")
    assert_equal(oecd.eissn, "1609-7491")


def test_issn():
    assert_equal(sofwarex.issn, None)
    assert_equal(oecd.issn, "0255-0822")


def test_oaallowsauthorpaid():
    assert_equal(sofwarex.oaallowsauthorpaid, None)
    assert_equal(oecd.oaallowsauthorpaid, None)


def test_openaccess():
    assert_equal(sofwarex.openaccess, 1)
    assert_equal(oecd.openaccess, None)


def test_openaccessstartdate():
    assert_equal(sofwarex.openaccessstartdate, None)
    assert_equal(oecd.openaccessstartdate, None)


def test_openaccesstype():
    assert_equal(sofwarex.openaccesstype, None)
    assert_equal(oecd.openaccesstype, None)


def test_openaccessarticle():
    assert_equal(sofwarex.openaccessarticle, True)
    assert_equal(oecd.openaccessarticle, None)


def test_openarchivearticle():
    assert_equal(sofwarex.openarchivearticle, None)
    assert_equal(oecd.openarchivearticle, None)


def test_openaccesssponsorname():
    assert_equal(sofwarex.openaccesssponsorname, None)
    assert_equal(oecd.openaccesssponsorname, None)


def test_openaccessuserlicense():
    assert_equal(sofwarex.openaccessuserlicense, None)
    assert_equal(oecd.openaccessuserlicense, None)


def test_publisher():
    assert_equal(sofwarex.publisher, "Elsevier BV")
    assert_equal(oecd.publisher, "OECD")


def test_scopus_source_link():
    expected1 = "https://www.scopus.com/source/sourceInfo.url?sourceId=21100422153"
    assert_equal(sofwarex.scopus_source_link, expected1)
    expected2 = "https://www.scopus.com/source/sourceInfo.url?sourceId=24107"
    assert_equal(oecd.scopus_source_link, expected2)


def test_self_link():
    expected1 = "https://api.elsevier.com/content/serial/title/issn/23527110"
    assert_equal(sofwarex.self_link, expected1)
    expected2 = "https://api.elsevier.com/content/serial/title/issn/02550822"
    assert_equal(oecd.self_link, expected2)

def test_sjrlist():
    assert_equal(sofwarex.sjrlist, [(2019, 0.445), (2020, 0.528)])
    assert_equal(oecd.sjrlist, [(1999, 2.723)])


def test_sniplist():
    assert_equal(sofwarex.sniplist, [(2019, 1.013), (2020, 1.093)])
    assert_equal(oecd.sniplist, None)


def test_source_id():
    assert_equal(sofwarex.source_id, 21100422153)
    assert_equal(oecd.source_id, 24107)


def test_subject_area():
    area = namedtuple('Subjectarea', 'area abbreviation code')
    expected1 = [
        area(area='Software', abbreviation='COMP', code=1712),
        area(area='Computer Science Applications', abbreviation='COMP', code=1706)
    ]
    assert_equal(sofwarex.subject_area, expected1)
    expected2 = [
        area(area='Geography, Planning and Development', abbreviation='SOCI', code=3305)
    ]
    assert_equal(oecd.subject_area, expected2)


def test_title():
    assert_equal(sofwarex.title, "SoftwareX")
    assert_equal(oecd.title, "OECD Economic Studies")



"""Superclass to access all Scopus retrieval APIs and dump the results."""

from typing import Union

from pybliometrics.scopus.superclasses import Base
from pybliometrics.scopus.utils import get_folder, URLS


class Retrieval(Base):
    def __init__(self,
                 identifier: Union[int, str],
                 api: str,
                 id_type: str = None,
                 **kwds: str
                 ) -> None:
        """Class intended as superclass to perform retrievals.

        :param identifier: The ID to look for.
        :param api: The name of the Scopus API to be accessed.  Allowed values:
                    AbstractRetrieval, AuthorRetrieval, CitationOverview,
                    AffiliationRetrieval.
        :param id_type: The type of the used ID.  Will only take effect for
                        the Abstract Retrieval API.
        :param kwds: Keywords passed on to requests header.  Must contain
                     fields and values specified in the respective
                     API specification.

        Raises
        ------
        KeyError
            If parameter `api` is not one of the allowed values.
        """
        # Construct URL and cache file name
        url = URLS[api]
        if api in ("AbstractRetrieval", "PlumXMetrics"):
            url += id_type + "/"
        if api == 'CitationOverview':
            stem = identifier.replace("/", "")
            if self._citation:
                identifier += "-" + self._citation
        else:
            url += identifier
            stem = identifier.replace('/', '_')
        self._cache_file_path = get_folder(api, self._view)/stem

        # Parse file contents
        params = {'view': self._view, **kwds}
        Base.__init__(self, params=params, url=url, api=api)



from collections import namedtuple
from typing import Dict, List, NamedTuple, Optional, Tuple, Union

from pybliometrics.scopus.superclasses import Search
from pybliometrics.scopus.utils import chained_get, make_search_summary


class SubjectClassifications(Search):
    @property
    def results(self) -> Optional[List[NamedTuple]]:
        """A list of namedtuples representing results of subject
        classifications search in the form (code, description, detail, abbrev).
        """
        out = []
        path = ['subject-classifications', 'subject-classification']
        search_results = chained_get(self._json, path, [])
        subj = namedtuple('Subject', self.fields)
        if isinstance(search_results, dict):
            for field_name in self.fields:
                if field_name not in search_results:
                    search_results[field_name] = None
            out.append(subj(**search_results))
        else:
            for result in search_results:
                missing_fields = set(self.fields).difference(result.keys())
                if missing_fields:
                    for field_name in missing_fields:
                        result[field_name] = None
                out.append(subj(**result))
        return out or None

    def __init__(self,
                 query: Dict,
                 refresh: Union[bool, int] = False,
                 fields: Union[List[str], Tuple[str, ...]] = None,
                 **kwds: str
                 ) -> None:
        """Interaction with the Subject Classifications Scopus API.

        :param query: Query parameters and corresponding fields. Allowed keys
                      'code', 'abbrev', 'description', 'detail'. For more
                      details on search fields please refer to
                      https://dev.elsevier.com/documentation/SubjectClassificationsAPI.wadl#d1e199.
        :param refresh: Whether to refresh the cached file if it exists or not.
                        If int is passed, cached file will be refreshed if the
                        number of days since last modification exceeds that value.
        :param fields: The fields to return when calling search results.
                       Allowed values: 'code', 'abbrev', 'description',
                       'detail'.  For details see
                       https://dev.elsevier.com/documentation/SubjectClassificationsAPI.wadl#d1e199.
        :param kwds: Keywords passed on as query parameters.  Must contain
                     fields and values mentioned in the API specification at
                     https://dev.elsevier.com/documentation/SubjectClassificationsAPI.wadl.

        Raises
        ------
        TypeError
            If returned fields are not passed in an iterable container.

        ValueError
            If any of the parameters `fields`, `refresh` or `query` is not
            one of the allowed values.

        Notes
        -----
        The directory for cached results is `{path}/{fname}`,
        where `path` is specified in your configuration file, and `fname` is
        the md5-hashed version of `query` dict turned into string in format
        of 'key=value' delimited by '&'.
        """
        # Checks
        allowed_query_keys = ('code', 'description', 'detail', 'abbrev')
        invalid = [k for k in query.keys() if k not in allowed_query_keys]
        if invalid:
            raise ValueError(f'Query key(s) "{", ".join(invalid)}" invalid.')
        self.fields = fields or allowed_query_keys
        if fields:
            try:
                return_fields = [i for i in fields]
            except TypeError:
                print("Fields must be iterable")
                raise
            if not set(return_fields).issubset(allowed_query_keys):
                raise ValueError("Parameter 'fields' must be one of " +
                                 f"{', '.join(allowed_query_keys)}.")

        # Query
        query['field'] = ','.join(self.fields)
        self._refresh = refresh
        self._query = str(query)
        self._view = None
        Search.__init__(self, query=query, api='SubjectClassifications', **kwds)
        path = ['subject-classifications', 'subject-classification']
        self._n = len(chained_get(self._json, path, []))

    def __str__(self):
        """Print a summary string."""
        areas = [r.code for r in self.results]
        return make_search_summary(self, "subject area", areas)
